# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import gettext

from PyQt6.QtGui import QRegularExpressionValidator, QIcon, QColor, QBrush

from PyQt6.QtWidgets import (
    QWidget,
    QDialog,
    QVBoxLayout,
    QDialogButtonBox,
    QTabWidget,
    QTableView,
    QPushButton,
    QMessageBox,
    QFileDialog,
    QComboBox,
    QLineEdit,
    QFormLayout,
    QLabel,
)

from PyQt6.QtCore import (
    Qt,
    QAbstractTableModel,
    QModelIndex,
    QVariant,
    QRegularExpression,
)

from . import FileNameFilters
from . import widgets
from .. import utils
from .. import students


t = gettext.translation("eyegrade", utils.locale_dir(), fallback=True)
_ = t.gettext


class DialogStudentId(QDialog):
    """Dialog to change the student id.

    Example (replace `parent` by the parent widget):

    dialog = DialogStudentId(parent)
    id = dialog.exec()

    """

    def __init__(self, parent, ranked_students, student_listings):
        super().__init__(parent)
        self.student_listings = student_listings
        self.setWindowTitle(_("Change the student id"))
        layout = QFormLayout()
        self.setLayout(layout)
        self.combo = widgets.StudentComboBox(parent=self)
        self.combo.add_students(self._combine_ranked_students(ranked_students))
        self.combo.editTextChanged.connect(self._check_value)
        self.combo.currentIndexChanged.connect(self._check_value)
        new_student_button = QPushButton(
            QIcon(utils.resource_path("new_id.svg")), _("New student"), parent=self
        )
        new_student_button.clicked.connect(self._new_student)
        self.buttons = QDialogButtonBox(
            (
                QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
            )
        )
        self.buttons.addButton(
            new_student_button, QDialogButtonBox.ButtonRole.ActionRole
        )
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        layout.addRow(_("Student id:"), self.combo)
        layout.addRow(self.buttons)

    def exec(self):
        """Shows the dialog and waits until it is closed.

        Returns a student object with the option selected by the user.
        The return value is None if the user cancels the dialog.

        """
        result = super().exec()
        if result == QDialog.DialogCode.Accepted:
            return self.combo.current_student()
        else:
            return None

    def _new_student(self):
        dialog = NewStudentDialog(self.student_listings, parent=self)
        student = dialog.exec()
        if student is not None:
            self.combo.add_student(student, set_current=True)
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setFocus()
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)

    def _check_value(self, param):
        if self.combo.current_student() is not None:
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)
        else:
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)

    def _combine_ranked_students(self, ranked_students):
        """Combine the ranked students with the rest of students.

        The ranked students are shown first in the combo box.

        """
        combined = list(ranked_students)
        seen_ids = set(student.student_id for student in ranked_students)
        for student in self.student_listings.iter_students():
            if student.student_id not in seen_ids:
                combined.append(student)
        return combined


class NewStudentDialog(QDialog):
    """Dialog to ask for a new student.

    It returns a Student object with the data of the student.

    """

    _last_combo_value = None

    def __init__(self, student_listings, group_index=None, parent=None):
        super().__init__(parent)
        self.student_listings = student_listings
        self.group_index = group_index
        self.setMinimumWidth(300)
        self.setWindowTitle(_("Add a new student"))
        layout = QFormLayout()
        self.setLayout(layout)
        self.id_field = QLineEdit(self)
        self.id_field.setValidator(
            QRegularExpressionValidator(QRegularExpression(r"\d+"), self)
        )
        self.id_field.textEdited.connect(self._check_values)
        self.name_field = QLineEdit(self)
        self.surname_field = QLineEdit(self)
        self.full_name_field = QLineEdit(self)
        self.name_label = QLabel(_("Given name"))
        self.surname_label = QLabel(_("Surname"))
        self.full_name_label = QLabel(_("Full name"))
        self.email_field = QLineEdit(self)
        self.email_field.setValidator(
            QRegularExpressionValidator(QRegularExpression(students.re_email), self)
        )
        self.email_field.textEdited.connect(self._check_values)
        self.combo = QComboBox(parent=self)
        self.combo.addItem(_("Separate given name and surname"))
        self.combo.addItem(_("Full name in just one field"))
        self.combo.currentIndexChanged.connect(self._update_combo)
        self.group_combo = QComboBox(parent=self)
        for listing in self.student_listings[1:]:
            self.group_combo.addItem(listing.group.name)
        layout.addRow(self.combo)
        layout.addRow(_("Id number"), self.id_field)
        layout.addRow(self.name_label, self.name_field)
        layout.addRow(self.surname_label, self.surname_field)
        layout.addRow(self.full_name_label, self.full_name_field)
        layout.addRow(_("Email"), self.email_field)
        layout.addRow(_("Group"), self.group_combo)
        self.buttons = QDialogButtonBox(
            (
                QDialogButtonBox.StandardButton.Ok
                | QDialogButtonBox.StandardButton.Cancel
            )
        )
        layout.addRow(self.buttons)
        self.buttons.accepted.connect(self.accept)
        self.buttons.rejected.connect(self.reject)
        self._check_values()
        # Set the combo box
        if NewStudentDialog._last_combo_value is None:
            NewStudentDialog._last_combo_value = 0
        self.combo.setCurrentIndex(NewStudentDialog._last_combo_value)
        if self.group_index is not None:
            self.group_combo.setCurrentIndex(self.group_index)
            self.group_combo.setEnabled(False)
        self._update_combo(NewStudentDialog._last_combo_value)

    def exec(self):
        """Shows the dialog and waits until it is closed.

        Returns the text of the option selected by the user, or None if
        the dialog is cancelled.

        """
        result = super().exec()
        if result == QDialog.DialogCode.Accepted:
            NewStudentDialog._last_combo_value = self.combo.currentIndex()
            listing = self.student_listings[self.group_combo.currentIndex() + 1]
            email = self.email_field.text()
            if not email:
                email = None
            if self.combo.currentIndex() == 0:
                # First name, last name
                student = students.Student(
                    self.id_field.text(),
                    None,
                    self.name_field.text(),
                    self.surname_field.text(),
                    email,
                    group_id=listing.group.identifier,
                )
            else:
                # Full name
                student = students.Student(
                    self.id_field.text(),
                    self.full_name_field.text(),
                    None,
                    None,
                    email,
                    group_id=listing.group.identifier,
                )
            try:
                listing.add_students((student,))
            except students.DuplicateStudentIdException:
                QMessageBox.critical(
                    self,
                    _("Adding a new student"),
                    _(
                        "The student cannot be added: "
                        "a student with the same id is already in the list"
                    ),
                )
                student = None
        else:
            student = None
        return student

    def _check_values(self):
        if self.id_field.hasAcceptableInput() and (
            not self.email_field.text() or self.email_field.hasAcceptableInput()
        ):
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(True)
        else:
            self.buttons.button(QDialogButtonBox.StandardButton.Ok).setEnabled(False)

    def _update_combo(self, new_index):
        if new_index == 0:
            self.name_field.setEnabled(True)
            self.surname_field.setEnabled(True)
            self.full_name_field.setEnabled(False)
            self.name_label.setEnabled(True)
            self.surname_label.setEnabled(True)
            self.full_name_label.setEnabled(False)
            self.full_name_field.setText("")
        else:
            self.name_field.setEnabled(False)
            self.surname_field.setEnabled(False)
            self.full_name_field.setEnabled(True)
            self.name_label.setEnabled(False)
            self.surname_label.setEnabled(False)
            self.full_name_label.setEnabled(True)
            self.name_field.setText("")
            self.surname_field.setText("")


class DialogStudents(QDialog):
    """Dialog to list students."""

    def __init__(self, parent, student_listings):
        super().__init__(parent)
        self.setWindowTitle(_("Manage students and student groups"))
        main_layout = QVBoxLayout(self)
        self.setLayout(main_layout)
        tabs = StudentGroupsTabs(self, student_listings=student_listings)
        main_layout.addWidget(tabs)
        buttons = QDialogButtonBox(QDialogButtonBox.StandardButton.Ok)
        buttons.accepted.connect(self.accept)
        main_layout.addWidget(buttons)

    def exec(self):
        """Shows the dialog and waits until it is closed."""
        result = super().exec()
        if result == QDialog.DialogCode.Accepted:
            return True
        else:
            return False


class DialogPreviewStudents(QDialog):
    """Dialog to preview and adjust a just loaded list students."""

    def __init__(self, parent, student_list, column_map):
        super().__init__(parent)
        self.setWindowTitle(_("Preview the students to be loaded"))
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.table = PreviewWidget(student_list, column_map, parent=self)
        self.button_swap = QPushButton(_("Swap first/last names"), parent=self)
        self.button_take_first = QPushButton(
            _("Take first name as full name"), parent=self
        )
        self.button_take_last = QPushButton(
            _("Take last name as full name"), parent=self
        )
        self.button_remove = QPushButton(_("Remove duplicate students"), parent=self)
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self.button_accept = buttons.button(QDialogButtonBox.StandardButton.Ok)
        layout.addWidget(self.table)
        layout.addWidget(self.button_swap)
        layout.addWidget(self.button_take_first)
        layout.addWidget(self.button_take_last)
        layout.addWidget(self.button_remove)
        layout.addWidget(buttons)
        layout.setAlignment(self.button_swap, Qt.AlignmentFlag.AlignHCenter)
        layout.setAlignment(self.button_take_first, Qt.AlignmentFlag.AlignHCenter)
        layout.setAlignment(self.button_take_last, Qt.AlignmentFlag.AlignHCenter)
        layout.setAlignment(self.button_remove, Qt.AlignmentFlag.AlignHCenter)
        self.button_swap.clicked.connect(self.table.swap_names)
        self.button_take_first.clicked.connect(self._take_first_name)
        self.button_take_last.clicked.connect(self._take_last_name)
        self.button_remove.clicked.connect(self._remove_duplicates)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        if students.StudentColumn.FIRST_NAME not in column_map:
            self._disable_buttons()
        if any(s.is_duplicate for s in student_list):
            self.button_accept.setEnabled(False)
            self.button_remove.setEnabled(True)
        else:
            self.button_remove.setEnabled(False)

    def exec(self):
        """Shows the dialog and waits until it is closed."""
        result = super().exec()
        if result == QDialog.DialogCode.Accepted:
            return True
        else:
            return False

    def _disable_buttons(self):
        self.button_swap.setEnabled(False)
        self.button_take_first.setEnabled(False)
        self.button_take_last.setEnabled(False)

    def _take_first_name(self):
        self.table.to_full_name(students.StudentColumn.FIRST_NAME)
        self._disable_buttons()

    def _take_last_name(self):
        self.table.to_full_name(students.StudentColumn.LAST_NAME)
        self._disable_buttons()

    def _remove_duplicates(self):
        self.table.remove_duplicates()
        self.button_accept.setEnabled(True)
        self.button_remove.setEnabled(False)


class GroupNameDialog(QDialog):
    def __init__(self, parent=None, group_name=""):
        super().__init__(parent)
        if not group_name:
            window_title = _("Create a new group of students")
            message = _(
                "Please, enter the name of the new group of students "
                "you want to create:"
            )
        else:
            window_title = _("Rename the group of students")
            message = _("Please, enter the new name of the group of students")
        self.setWindowTitle(window_title)
        self.group_name = group_name
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self._name_widget = QLineEdit(self)
        if group_name:
            self._name_widget.setText(group_name)
            self._name_widget.selectAll()
        self._name_widget.textChanged.connect(self._group_name_changed)
        main_line = widgets.LineContainer(
            self, QLabel(_("Group name")), self._name_widget
        )
        buttons = QDialogButtonBox(
            QDialogButtonBox.StandardButton.Ok | QDialogButtonBox.StandardButton.Cancel
        )
        self._button_ok = buttons.button(QDialogButtonBox.StandardButton.Ok)
        layout.addWidget(QLabel(message))
        layout.addWidget(main_line)
        layout.addWidget(buttons)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        self._group_name_changed(group_name)

    def exec(self):
        result = super().exec()
        if result == QDialog.DialogCode.Accepted:
            return self._name_widget.text()
        else:
            return None

    def _group_name_changed(self, group_name):
        if group_name:
            self._button_ok.setEnabled(True)
        else:
            self._button_ok.setEnabled(False)


class StudentGroupsTabs(QWidget):
    def __init__(self, parent, student_listings=None):
        super().__init__(parent)
        if student_listings is None:
            self.student_listings = students.StudentListings()
            inserted_group = students.StudentGroup(0, "INSERTED")
            self.student_listings.create_listing(inserted_group)
        else:
            self.student_listings = student_listings
        layout = QVBoxLayout(self)
        self.setLayout(layout)
        self.tabs = QTabWidget(self)
        # Special tab for creating a new group:
        self.tabs.addTab(QWidget(), "  +  ")
        # Group 0 is special: don't show it
        for listing in self.student_listings[1:]:
            self._add_group_tab(listing)
        # At least one normal group needs to be present:
        if len(self.student_listings) == 1:
            self._create_default_group()
        button_load = QPushButton(_("Add students from file"), parent=self)
        button_new_student = QPushButton(
            QIcon(utils.resource_path("new_id.svg")), _("New student"), parent=self
        )
        button_remove = QPushButton(
            QIcon(utils.resource_path("discard.svg")), _("Remove group"), parent=self
        )
        layout.addWidget(self.tabs)
        layout.addWidget(button_load)
        layout.addWidget(button_new_student)
        layout.addWidget(button_remove)
        layout.setAlignment(button_load, Qt.AlignmentFlag.AlignHCenter)
        layout.setAlignment(button_new_student, Qt.AlignmentFlag.AlignHCenter)
        layout.setAlignment(button_remove, Qt.AlignmentFlag.AlignHCenter)
        self.tabs.setCurrentIndex(0)
        self._active_tab = 0
        self.tabs.currentChanged.connect(self._tab_changed)
        self.tabs.tabBarDoubleClicked.connect(self._rename_group)
        button_load.clicked.connect(self._load_students)
        button_new_student.clicked.connect(self._new_student)
        button_remove.clicked.connect(self._remove_group)

    def _load_students(self):
        index = self.tabs.currentIndex()
        file_name, __ = QFileDialog.getOpenFileName(
            self,
            _("Select the student list file"),
            "",
            FileNameFilters.student_list,
            None,
            QFileDialog.Option.DontUseNativeDialog,
        )
        try:
            if file_name:
                with students.StudentReader.create(file_name) as reader:
                    student_list = list(reader.students())
                    column_map = reader.column_map.normalize()
                # Flag duplicate student ids:
                warn_duplicates = False
                for s in self.student_listings.find_duplicates(student_list):
                    s.is_duplicate = True
                    warn_duplicates = True
                if warn_duplicates:
                    QMessageBox.warning(
                        self,
                        _("Importing a student list"),
                        _(
                            "Some student ids are already in the list. "
                            "Remove them or cancel the import operation."
                        ),
                    )
                preview_dialog = DialogPreviewStudents(self, student_list, column_map)
                result = preview_dialog.exec()
                if result == QMessageBox.DialogCode.Accepted:
                    self.tabs.widget(index).add_students(student_list)
        except Exception as e:
            QMessageBox.critical(
                self, _("Error in student list"), file_name + "\n\n" + str(e)
            )

    def _new_student(self):
        index = self.tabs.currentIndex()
        dialog = NewStudentDialog(self.student_listings, group_index=index, parent=self)
        student = dialog.exec()
        if student is not None:
            self.tabs.widget(index).listing_updated()

    def _remove_group(self):
        index = self.tabs.currentIndex()
        if len(self.student_listings[index + 1]) > 0:
            result = QMessageBox.warning(
                self,
                _("Warning"),
                _(
                    "This group and its students will be removed. "
                    "Are you sure you want to continue?"
                ),
                QMessageBox.StandardButton.Yes | QMessageBox.StandardButton.No,
                QMessageBox.StandardButton.No,
            )
            remove = result == QMessageBox.StandardButton.Yes
        else:
            remove = True
        if remove:
            try:
                self.student_listings.remove_at(index + 1)
                if len(self.student_listings) > 1:
                    if index == self.tabs.count() - 2:
                        self.tabs.setCurrentIndex(index - 1)
                    else:
                        self.tabs.setCurrentIndex(index + 1)
                else:
                    self._create_default_group()
                    self.tabs.setCurrentIndex(1)
                self.tabs.removeTab(index)
            except students.CantRemoveGroupException:
                QMessageBox.critical(
                    self,
                    _("Error"),
                    _(
                        "This group cannot be removed because "
                        "exams have been graded for some of its students."
                    ),
                )

    def _add_group_tab(self, listing, show=False):
        self.tabs.insertTab(
            self.tabs.count() - 1, GroupWidget(listing, self), listing.group.name
        )
        if show:
            self.tabs.setCurrentIndex(self.tabs.count() - 2)

    def _new_group(self):
        group_name = GroupNameDialog(parent=self).exec()
        if group_name is not None:
            group = students.StudentGroup(None, group_name)
            listing = self.student_listings.create_listing(group)
            self._add_group_tab(listing, show=True)
        else:
            self.tabs.setCurrentIndex(self._active_tab)

    def _create_default_group(self):
        group = students.StudentGroup(None, _("Students"))
        listing = self.student_listings.create_listing(group)
        self._add_group_tab(listing, show=True)

    def _rename_group(self, index):
        name = self.student_listings[index + 1].group.name
        new_name = GroupNameDialog(group_name=name, parent=self).exec()
        if new_name is not None and new_name != name:
            self.student_listings[index + 1].rename(new_name)
            self.tabs.setTabText(index, new_name)

    def _tab_changed(self, index):
        if index == self.tabs.count() - 1:
            # The last (special) tab has been activated: create a new group
            self._new_group()
        self._active_tab = self.tabs.currentIndex()


class GroupWidget(QWidget):
    _COLUMN_MAP = students.StudentColumnMap(
        columns=[
            students.StudentColumn.SEQUENCE_NUM,
            students.StudentColumn.ID,
            students.StudentColumn.NAME,
        ]
    )

    def __init__(self, listing, student_tabs):
        super().__init__(student_tabs.tabs)
        self.listing = listing
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.table = QTableView()
        self.table.setMinimumWidth(500)
        self.table.setMinimumHeight(300)
        layout.addWidget(self.table)
        self.model = StudentsTableModel(listing, GroupWidget._COLUMN_MAP, self)
        self.table.setModel(self.model)
        self.table.setSelectionBehavior(QTableView.SelectionBehavior.SelectRows)
        layout.setAlignment(self.table, Qt.AlignmentFlag.AlignHCenter)
        self._resize_table()

    def add_students(self, student_list):
        self.listing.add_students(student_list)
        self.listing_updated()

    def listing_updated(self):
        self.model.data_reset()
        self._resize_table()

    def _resize_table(self):
        self.table.resizeColumnToContents(0)
        self.table.horizontalHeader().setStretchLastSection(True)


class PreviewWidget(QWidget):
    def __init__(self, student_list, column_map, parent=None):
        super().__init__(parent)
        self.listing = students.GroupListing(None, student_list)
        self.column_map = column_map
        layout = QVBoxLayout()
        self.setLayout(layout)
        self.table = QTableView()
        self.table.setMinimumWidth(600)
        self.table.setMinimumHeight(300)
        layout.addWidget(self.table)
        self.model = StudentsTableModel(self.listing, column_map, self)
        self.table.setModel(self.model)
        self.table.setSelectionMode(QTableView.SelectionMode.NoSelection)
        layout.setAlignment(self.table, Qt.AlignmentFlag.AlignHCenter)
        self._resize_table()

    def swap_names(self):
        for s in self.listing.students:
            s.first_name, s.last_name = s.last_name, s.first_name
        self.model.data_reset()
        self._resize_table()

    def to_full_name(self, column):
        attr_name = students.ATTR_NAME[column]
        for s in self.listing.students:
            s.full_name = getattr(s, attr_name)
            s.first_name = ""
            s.last_name = ""
        self.column_map = self.column_map.to_full_name()
        self.model.data_reset(column_map=self.column_map)
        self._resize_table()

    def remove_duplicates(self):
        self.listing.remove_students(
            [s for s in self.listing.students if s.is_duplicate]
        )
        self.model.data_reset()

    def _resize_table(self):
        for i in range(len(self.column_map) - 1):
            self.table.resizeColumnToContents(i)
        self.table.horizontalHeader().setStretchLastSection(True)


class StudentsTableModel(QAbstractTableModel):
    """Table for showing a student list."""

    _headers = {
        students.StudentColumn.SEQUENCE_NUM: "#",
        students.StudentColumn.ID: _("Id"),
        students.StudentColumn.NAME: _("Name"),
        students.StudentColumn.FIRST_NAME: _("First name"),
        students.StudentColumn.LAST_NAME: _("Last name"),
        students.StudentColumn.FULL_NAME: _("Name"),
        students.StudentColumn.EMAIL: _("Email"),
    }

    _alignment = {
        students.StudentColumn.SEQUENCE_NUM: Qt.AlignmentFlag.AlignRight,
        students.StudentColumn.ID: Qt.AlignmentFlag.AlignRight,
        students.StudentColumn.NAME: Qt.AlignmentFlag.AlignLeft,
        students.StudentColumn.FIRST_NAME: Qt.AlignmentFlag.AlignLeft,
        students.StudentColumn.LAST_NAME: Qt.AlignmentFlag.AlignLeft,
        students.StudentColumn.FULL_NAME: Qt.AlignmentFlag.AlignLeft,
        students.StudentColumn.EMAIL: Qt.AlignmentFlag.AlignLeft,
    }

    def __init__(self, listing, column_map, parent=None):
        super().__init__(parent=None)
        self.column_map = column_map
        self.data_reset(listing=listing)

    def data_reset(self, listing=None, column_map=None):
        self.beginResetModel()
        if listing is not None:
            self.listing = listing
        if column_map is not None:
            self.column_map = column_map
        self.endResetModel()

    def rowCount(self, parent=QModelIndex()):
        return len(self.listing)

    def columnCount(self, parent=QModelIndex()):
        # Columns: sequence num, id, full name
        return len(self.column_map)

    def headerData(self, section, orientation, role):
        if role == Qt.ItemDataRole.DisplayRole:
            if orientation == Qt.Orientation.Horizontal:
                return StudentsTableModel._headers[self.column_map[section]]
            else:
                return QVariant()
        else:
            return QVariant()

    def data(self, index, role=Qt.ItemDataRole.DisplayRole):
        column = index.column()
        student = self.listing[index.row()]
        if role == Qt.ItemDataRole.DisplayRole:
            return self.column_map.data(column, student)
        elif role == Qt.ItemDataRole.TextAlignmentRole:
            return StudentsTableModel._alignment[self.column_map[column]]
        elif role == Qt.ItemDataRole.BackgroundRole:
            if student.is_duplicate:
                return QBrush(QColor(255, 165, 165))
            else:
                return QVariant()
        else:
            return QVariant()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import sqlite3
import os
from typing import Optional

from . import utils
from . import scoring
from . import exams
from . import students
from . import capture
from . import images
from . import export


class SessionDB:
    """Access to a session SQLite database.

    This class encapsulates access functions to the session database.

    """

    DB_SCHEMA_VERSION = 6
    COMPATIBLE_SCHEMAS = (1, 2, 3, 4, 5, 6)

    GRADING_MODE_ONE_CORRECT = 1
    GRADING_MODE_MULTI_CORRECT = 2

    _table_session = """
        CREATE TABLE Session (
            db_schema_version INTEGER,
            eyegrade_version STRING,
            title TEXT,
            description TEXT,
            dimensions TEXT NOT NULL,
            grading_mode INTEGER NOT NULL,
            scores_mode INTEGER NOT NULL,
            base_score_correct TEXT,
            base_score_incorrect TEXT,
            base_score_blank TEXT,
            id_num_digits INTEGER NOT NULL,
            survey_mode INTEGER NOT NULL,
            left_to_right_numbering INTEGER NOT NULL,
            capture_pattern TEXT NOT NULL
        )"""

    _table_questions = """
        CREATE TABLE Questions (
            model INTEGER NOT NULL,
            question INTEGER NOT NULL,
            permutation TEXT,
            variation INTEGER,
            score_correct TEXT,
            score_incorrect TEXT,
            score_blank TEXT,
            score_weight TEXT
        )"""

    _table_solutions = """
        CREATE TABLE Solutions (
            model INTEGER NOT NULL,
            question INTEGER NOT NULL,
            solution INTEGER
        )"""

    _table_models = """
        CREATE TABLE Models (
            model INTEGER NOT NULL,
            model_variation INTEGER
    )"""

    _table_exams = """
        CREATE TABLE Exams (
            exam_id INTEGER PRIMARY KEY NOT NULL,
            student INTEGER,
            model INTEGER,
            correct INTEGER,
            incorrect INTEGER,
            blank INTEGER,
            score REAL,
            FOREIGN KEY(student) REFERENCES Students(db_id)
        )"""

    _table_students = """
        CREATE TABLE Students (
            db_id INTEGER PRIMARY KEY NOT NULL,
            student_id TEXT,
            full_name TEXT,
            first_name TEXT,
            last_name TEXT,
            email TEXT,
            group_id INTEGER NOT NULL,
            sequence_num INTEGER NOT NULL,
            FOREIGN KEY(group_id) REFERENCES StudentGroups(group_id)
        )"""

    _table_student_groups = """
        CREATE TABLE StudentGroups (
            group_id INTEGER PRIMARY KEY NOT NULL,
            group_name TEXT NOT NULL
        )"""

    _table_answers = """
        CREATE TABLE Answers (
            exam_id INTEGER NOT NULL,
            question INTEGER NOT NULL,
            answer INTEGER NOT NULL,
            FOREIGN KEY(exam_id) REFERENCES Exams(exam_id)
        )"""

    _table_answer_cells = """
        CREATE TABLE AnswerCells (
            exam_id INTEGER NOT NULL,
            question INTEGER NOT NULL,
            choice INTEGER NOT NULL,
            center_x INTEGER NOT NULL,
            center_y INTEGER NOT NULL,
            diagonal INTEGER NOT NULL,
            lux INTEGER,
            luy INTEGER,
            rux INTEGER,
            ruy INTEGER,
            ldx INTEGER,
            ldy INTEGER,
            rdx INTEGER,
            rdy INTEGER,
            FOREIGN KEY(exam_id) REFERENCES Exams(exam_id)
        )"""

    _table_id_cells = """
        CREATE TABLE IdCells (
            exam_id INTEGER NOT NULL,
            digit INTEGER NOT NULL,
            lux INTEGER NOT NULL,
            luy INTEGER NOT NULL,
            rux INTEGER NOT NULL,
            ruy INTEGER NOT NULL,
            ldx INTEGER NOT NULL,
            ldy INTEGER NOT NULL,
            rdx INTEGER NOT NULL,
            rdy INTEGER NOT NULL,
            FOREIGN KEY(exam_id) REFERENCES Exams(exam_id)
        )"""

    _index_student_id = """
        CREATE UNIQUE INDEX idx_student_id ON Students(student_id)"""

    def __init__(self, session_file, open=True):
        """Opens a session database.

        For opening an existing session, just pass as `session_file` the
        session DB file name or the session's directory name.

        """
        if os.path.isdir(session_file):
            self.session_dir = session_file
        else:
            self.session_dir = os.path.dirname(session_file)
        self._check_session_directory()
        if open:
            self._open()
        else:
            self.conn = None

    def _open(self):
        db_file = os.path.join(self.session_dir, "session.eyedb")
        self.conn = sqlite3.connect(db_file)
        self.conn.row_factory = sqlite3.Row
        self._enable_foreign_key_constrains()
        self.schema_version = self._check_schema()
        self.exam_config = self._load_exam_config()
        self._compute_num_questions_and_choices()
        self.capture_save_func = lambda name: None
        self._student_listings = None

    @property
    def student_listings(self):
        if self._student_listings is None:
            groups = self.get_student_groups(ignore_empty_groups=False)
            self._student_listings = StudentListingsFromDB(self)
            for group in groups:
                self._student_listings.add_listing(self.get_group_listing(group))
        return self._student_listings

    def close(self):
        self.conn.close()

    def __enter__(self):
        if self.conn is None:
            self._open()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    def store_exam(self, exam_id, exam_capture, decisions, score, store_captures=True):
        student_db_id = self._student_db_id(decisions.student)
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT INTO Exams VALUES " "(?, ?, ?, ?, ?, ?, ?)",
            (
                exam_id,
                student_db_id,
                _enc_model(decisions.model),
                score.correct,
                score.incorrect,
                score.blank,
                score.score,
            ),
        )
        if decisions.answers is not None:
            self._store_answers(exam_id, decisions.answers, commit=False)
            self._store_answer_cells(exam_id, exam_capture.answer_cells, commit=False)
        if exam_capture.id_cells:
            self._store_id_cells(exam_id, exam_capture.id_cells, commit=False)
        self.conn.commit()
        if store_captures:
            self.save_raw_capture(exam_id, exam_capture)
            self.save_drawn_capture(exam_id, exam_capture, decisions.student)

    def remove_exam(self, exam_id):
        cursor = self.conn.cursor()
        student = self._read_student_by_exam(exam_id)
        cursor.execute("DELETE FROM Answers WHERE exam_id=?", (exam_id,))
        cursor.execute("DELETE FROM AnswerCells WHERE exam_id=?", (exam_id,))
        cursor.execute("DELETE FROM IdCells WHERE exam_id=?", (exam_id,))
        cursor.execute("DELETE FROM Exams WHERE exam_id=?", (exam_id,))
        self.conn.commit()
        self.remove_drawn_capture(exam_id, student)
        self.remove_raw_capture(exam_id)

    def update_answer(
        self, exam_id, question, exam_capture, decisions, score, store_captures=True
    ):
        new_answer = decisions.answers[question]
        self._update_answer(exam_id, question, new_answer, commit=False)
        self._update_score(exam_id, score, commit=False)
        self.conn.commit()
        if store_captures:
            self.save_drawn_capture(exam_id, exam_capture, decisions.student)

    def update_score(self, exam: exams.Exam, commit: bool = True):
        self._update_score(exam.exam_id, exam.score, commit=commit)

    def update_exam_config_scores(self, exam_data, commit=True):
        # Store base scores
        base_scores = _base_scores(exam_data)
        cursor = self.conn.cursor()
        cursor.execute(
            """UPDATE Session
               SET
               scores_mode=?,
               base_score_correct=?,
               base_score_incorrect=?,
               base_score_blank=?
            """,
            (exam_data.scores_mode, base_scores[0], base_scores[1], base_scores[2]),
        )
        # Store question scores
        for model in exam_data.models:
            model_code = _enc_model(model)
            scores_c, scores_i, scores_b, weights = _question_scores(exam_data, model)
            data = zip(scores_c, scores_i, scores_b, weights)
            for i, scores in enumerate(data):
                cursor.execute(
                    """
                        UPDATE Questions
                        SET
                        score_correct=?,
                        score_incorrect=?,
                        score_blank=?,
                        score_weight=?
                        WHERE model=? AND question=?
                    """,
                    (*scores, model_code, i),
                )
        if commit:
            self.conn.commit()

    def update_student(self, exam_id, exam_capture, decisions, store_captures=True):
        new_student_db_id = self._student_db_id(decisions.student)
        old_student = self._read_student_by_exam(exam_id)
        cursor = self.conn.cursor()
        cursor.execute(
            "UPDATE Exams SET student = ? WHERE exam_id = ?",
            (new_student_db_id, exam_id),
        )
        self.conn.commit()
        self.remove_drawn_capture(exam_id, old_student)
        if store_captures:
            self.save_drawn_capture(exam_id, exam_capture, decisions.student)

    def _store_new_student(self, student):
        if student.group_id is None:
            student.group_id = 0
        listing = self.student_listings.listing_by_group_id(student.group_id)
        listing.add_students((student,))

    def store_students(self, student_list, commit=True):
        cursor = self.conn.cursor()
        data = []
        # Insert one by one with execute() instead of using executemany()
        # to get the autogenerated ids
        for student in student_list:
            if student.group_id is None:
                raise ValueError("Students must belong to a group")
            data = [
                student.student_id,
                student.full_name,
                student.first_name,
                student.last_name,
                student.email,
                student.group_id,
                student.sequence_num,
            ]
            if self.schema_version >= 2:
                cursor.execute(
                    "INSERT INTO Students "
                    "(student_id, full_name, first_name, "
                    " last_name, email, group_id, sequence_num) "
                    "VALUES (?, ?, ?, ?, ?, ?, ?)",
                    data,
                )
            else:
                del data[2:4]
                cursor.execute(
                    "INSERT INTO Students "
                    "(student_id, name, email, group_id, sequence_num) "
                    "VALUES (?, ?, ?, ?, ?)",
                    data,
                )
            student.db_id = cursor.lastrowid
            student.is_in_database = True
        if commit:
            self.conn.commit()

    def get_students(self, group_id=None):
        cursor = self.conn.cursor()
        if group_id is None:
            cursor.execute("SELECT * FROM Students ORDER BY sequence_num")
        else:
            cursor.execute(
                "SELECT * FROM Students WHERE group_id=? " "ORDER BY sequence_num",
                (group_id,),
            )
        return [self._student_from_row(row) for row in cursor]

    def get_student_groups(self, ignore_empty_groups=True):
        """Return the list of student groups.

        If `ignore_empty_groups` is set, only the groups with at least one
        student are returned.

        The groups are returned as a list of students.StudentGroup objects.

        """
        groups = []
        cursor = self.conn.cursor()
        if ignore_empty_groups:
            query = (
                "SELECT StudentGroups.group_id, group_name "
                "FROM StudentGroups "
                "INNER JOIN Students "
                "ON Students.group_id=StudentGroups.group_id "
                "GROUP BY Students.group_id"
            )
        else:
            query = "SELECT group_id, group_name " "FROM StudentGroups"
        for row in cursor.execute(query):
            # Use index instead of name because of incompatibilities
            # in the keys between older and newer versions of python/sql:
            groups.append(students.StudentGroup(row[0], row[1]))
        return groups

    def get_group_listing(self, group):
        """Return a `student.GroupListing` object."""
        return GroupListingFromDB(
            self, group, self.get_students(group_id=group.identifier)
        )

    def num_exams_from_group(self, group):
        """Get the number of exams of students in the given group."""
        cursor = self.conn.cursor()
        cursor.execute(
            (
                "SELECT COUNT(*) "
                "FROM Exams "
                "INNER JOIN Students ON student=db_id "
                "INNER JOIN StudentGroups "
                "    ON Students.group_id=StudentGroups.group_id "
                "WHERE Students.group_id=?"
            ),
            (group.identifier,),
        )
        return cursor.fetchone()[0]

    def remove_group(self, group):
        """Removes a group and all its students."""
        cursor = self.conn.cursor()
        cursor.execute("DELETE FROM Students WHERE group_id=?", (group.identifier,))
        cursor.execute(
            "DELETE FROM StudentGroups WHERE group_id=?", (group.identifier,)
        )
        self.conn.commit()

    def create_student_group(self, group_name):
        cursor = self.conn.cursor()
        cursor.execute(
            "INSERT INTO StudentGroups (group_name) VALUES (?)", (group_name,)
        )
        group_id = cursor.lastrowid
        self.conn.commit()
        return students.StudentGroup(group_id, group_name)

    def rename_student_group(self, group):
        """The new name is taken from the group input parameter."""
        cursor = self.conn.cursor()
        cursor.execute(
            "UPDATE StudentGroups SET group_name=? WHERE group_id=?",
            (group.name, group.identifier),
        )
        self.conn.commit()

    def next_exam_id(self):
        cursor = self.conn.cursor()
        cursor.execute("SELECT MAX(exam_id) FROM Exams")
        result = cursor.fetchone()[0]
        if result is not None:
            return int(result) + 1
        else:
            return 1

    def save_legacy_answers(self):
        file_name = os.path.join(self.session_dir, "eyegrade-answers.csv")
        file_format = export.FileFormat.CSV_TABS
        with export.create_writer(file_name, file_format) as writer:
            for exam in self.exams_iterator():
                data = [
                    exam["exam_id"],
                    exam["student_id"] if exam["student_id"] else -1,
                    exam["model"],
                    exam["correct"],
                    exam["incorrect"],
                    exam["score"] if exam["score"] is not None else "?",
                    "/".join([str(answer) for answer in exam["answers"]]),
                ]
                writer.append_row(data)

    def export_grades(self, export_helper):
        if export_helper.add_column_headers:
            column_headers = export_helper.column_headers()
        with export_helper.create_writer() as writer:
            for i, (group, title) in enumerate(export_helper.iter_groups()):
                if i > 0:
                    writer.append_sheet()
                writer.set_sheet_title(title)
                if export_helper.add_column_headers:
                    writer.append_row(column_headers)
                if export_helper.export_by_exam:
                    iterator = self.exams_iterator()
                else:
                    iterator = self.grades_iterator(
                        all_students=export_helper.all_students,
                        sort_key=export_helper.sort_by,
                        student_group=group,
                    )
                for exam in iterator:
                    writer.append_row(export_helper.data(exam))

    def exams_iterator(self):
        cursor = self.conn.cursor()
        for row in cursor.execute(
            "SELECT "
            "exam_id, student_id, model, "
            "correct, incorrect, score "
            "FROM Exams "
            "LEFT JOIN Students ON student = db_id "
            "ORDER BY exam_id"
        ):
            exam = dict(row)
            exam["model"] = _dec_model(exam["model"])
            exam["answers"] = self.read_answers(exam["exam_id"])
            yield exam

    def _grades_iterator_query(self, all_students, sort_key, student_group):
        if all_students:
            join_type = "LEFT"
        else:
            join_type = "INNER"
        if sort_key == export.SortBy.STUDENT_LIST:
            sort_clause = "ORDER BY group_id, sequence_num"
        elif sort_key == export.SortBy.LAST_NAME:
            if self.schema_version >= 2:
                sort_clause = "ORDER BY last_name, group_id, sequence_num"
            else:
                sort_clause = "ORDER BY name, group_id, sequence_num"
        elif sort_key == export.SortBy.GRADING_SEQUENCE:
            sort_clause = "ORDER BY exam_id"
        if student_group is not None:
            where_clause = "WHERE group_id = {0.identifier} ".format(student_group)
        else:
            where_clause = ""
        return (
            "SELECT "
            "* "
            "FROM Students "
            "{0} JOIN Exams ON student = db_id "
            "{1}"
            "{2}"
        ).format(join_type, where_clause, sort_clause)

    def grades_iterator(
        self,
        all_students=True,
        sort_key=export.SortBy.STUDENT_LIST,
        student_group=None,
        by_exams=False,
    ):
        if by_exams:
            query = (
                "SELECT "
                "* "
                "FROM Exams "
                "LEFT JOIN Students ON student = db_id "
                "ORDER BY exam_id"
            )
        else:
            query = self._grades_iterator_query(all_students, sort_key, student_group)
        cursor = self.conn.cursor()
        for row in cursor.execute(query):
            student = self._student_from_row(row)
            exam = {"student": student}
            for key in ("exam_id", "model", "correct", "incorrect", "score"):
                exam[key] = row[key]
            exam["model"] = _dec_model(exam["model"])
            exam["answers"] = self.read_answers(exam["exam_id"])
            for key, value in exam.items():
                if value is None:
                    exam[key] = ""
            yield exam

    def read_answers(self, exam_id):
        answers = [0] * self.exam_config.num_questions
        cursor = self.conn.cursor()
        for row in cursor.execute(
            "SELECT question, answer FROM Answers " "WHERE exam_id = ?", (exam_id,)
        ):
            answers[row["question"]] = row["answer"]
        return answers

    def read_exams(self):
        cursor = self.conn.cursor()
        exam_list = []
        students_rank = self.student_listings.sorted_students()
        for row in cursor.execute(
            "SELECT "
            "exam_id, student_id, model, "
            "correct, incorrect, blank, score "
            "FROM Exams "
            "LEFT JOIN Students ON student = db_id"
        ):
            exam = ExamFromDB(row, students_rank, self)
            exam_list.append(exam)
        return exam_list

    def read_capture(self, exam_id):
        image = self.load_raw_capture(exam_id)
        answer_cells = self._read_answer_cells(exam_id)
        id_cells = self._read_id_cells(exam_id)
        return capture.ExamCapture(image, answer_cells, id_cells)

    def _read_answer_cells(self, exam_id):
        all_cells = []
        question_cells = []
        last_question_num = None
        cursor = self.conn.cursor()
        for row in cursor.execute(
            "SELECT * FROM AnswerCells WHERE exam_id=? " "ORDER BY question, choice",
            (exam_id,),
        ):
            cell = _create_cell_from_row(row, is_id_cell=False)
            if last_question_num is None:
                last_question_num = row["question"]
            elif last_question_num != row["question"]:
                all_cells.append(question_cells)
                last_question_num = row["question"]
                question_cells = []
            question_cells.append(cell)
        all_cells.append(question_cells)
        return all_cells

    def _read_id_cells(self, exam_id):
        cells = []
        cursor = self.conn.cursor()
        for row in cursor.execute(
            "SELECT * FROM IdCells WHERE exam_id=? " "ORDER BY digit", (exam_id,)
        ):
            cells.append(_create_cell_from_row(row, is_id_cell=True))
        return cells

    def save_drawn_capture(self, exam_id, exam_capture, student, image_saver=None):
        name = utils.capture_name(self.exam_config.capture_pattern, exam_id, student)
        drawn_name = os.path.join(self.session_dir, "captures", name)
        if image_saver is not None:
            image_saver.save(drawn_name)
        elif self.capture_save_func is not None:
            self.capture_save_func(drawn_name)
        else:
            exam_capture.save_image_drawn(drawn_name)

    def save_raw_capture(self, exam_id, exam_capture):
        raw_name = os.path.join(
            self.session_dir, "internal", "raw-{0}.png".format(exam_id)
        )
        exam_capture.save_image_raw(raw_name)

    def load_raw_capture(self, exam_id):
        return images.load_image(self.get_raw_capture_path(exam_id))

    def get_raw_capture_path(self, exam_id):
        path = os.path.join(self.session_dir, "internal", "raw-{0}.png".format(exam_id))
        if not os.path.isfile(path):
            path = utils.resource_path("not_found.png")
        return path

    def remove_drawn_capture(self, exam_id, student):
        name = utils.capture_name(self.exam_config.capture_pattern, exam_id, student)
        drawn_name = os.path.join(self.session_dir, "captures", name)
        if os.path.exists(drawn_name):
            os.remove(drawn_name)

    def remove_raw_capture(self, exam_id):
        raw_name = os.path.join(
            self.session_dir, "internal", "raw-{0}.png".format(exam_id)
        )
        if os.path.exists(raw_name):
            os.remove(raw_name)

    def _check_schema(self):
        cursor = self.conn.cursor()
        cursor.execute("SELECT db_schema_version, eyegrade_version " "FROM Session")
        row = cursor.fetchone()
        schema = row["db_schema_version"]
        version = row["eyegrade_version"]
        if schema not in SessionDB.COMPATIBLE_SCHEMAS:
            raise utils.EyegradeException(
                "",
                key="incompatible_schema",
                format_params=(utils.program_name, utils.version, version),
            )
        return schema

    def _check_session_directory(self):
        db_file = os.path.join(self.session_dir, "session.eyedb")
        if not os.path.exists(db_file):
            raise utils.EyegradeException("", key="no_session_db")
        if not check_file_is_sqlite(db_file):
            raise utils.EyegradeException("", key="session_invalid")
        if not os.path.exists(
            os.path.join(self.session_dir, "captures")
        ) or not os.path.exists(os.path.join(self.session_dir, "internal")):
            raise utils.EyegradeException("", key="corrupt_session_dir")

    def _student_db_id(self, student):
        if student is not None:
            if not student.is_in_database:
                self._store_new_student(student)
            student_db_id = student.db_id
        else:
            student_db_id = None
        return student_db_id

    def _read_student_by_exam(self, exam_id):
        cursor = self.conn.cursor()
        cursor.execute(
            "SELECT * FROM Students "
            "INNER JOIN Exams ON Students.db_id = Exams.student "
            "WHERE exam_id = ?",
            (exam_id,),
        )
        row = cursor.fetchone()
        if row is not None:
            return self._student_from_row(row)
        else:
            return None

    def _read_cell_geometries(self, exam_id, load_corners=False):
        if not load_corners:
            query = (
                "SELECT question, choice, center_x, center_y, diagonal "
                "FROM CellGeometries WHERE exam_id = ?"
            )
        else:
            query = "SELECT * FROM CellGeometries WHERE exam_id = ?"
        cursor = self.conn.cursor()
        return [row for row in cursor.execute(query, exam_id)]

    def _compute_num_questions_and_choices(self):
        self.num_choices = []
        self.num_questions = 0
        dimensions = self.exam_config.dimensions
        for table in dimensions:
            self.num_choices.extend([table[0]] * table[1])
            self.num_questions += table[1]

    def _load_exam_config(self):
        self.exam_config = exams.ExamConfig()
        cursor = self.conn.cursor()
        cursor.execute("SELECT * FROM Session")
        row = cursor.fetchone()
        if row is None:
            raise utils.EyegradeException("", key="session_invalid")
        self.exam_config.set_dimensions(row["dimensions"])
        self.exam_config.id_num_digits = row["id_num_digits"]
        self.exam_config.survey_mode = True if row["survey_mode"] else False
        self.exam_config.left_to_right_numbering = (
            True if row["left_to_right_numbering"] else False
        )
        self.exam_config.capture_pattern = row["capture_pattern"]
        if self.schema_version >= 3:
            scores_mode = row["scores_mode"]
            self.exam_config.scores_mode = scores_mode
            if scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
                if row["base_score_correct"] is not None:
                    base_scores = scoring.QuestionScores(
                        row["base_score_correct"],
                        row["base_score_incorrect"],
                        row["base_score_blank"],
                    )
                    self.exam_config.set_base_scores(base_scores)
                else:
                    raise utils.EyegradeException("", key="session_invalid")
            else:
                if row["base_score_correct"] is not None:
                    raise utils.EyegradeException("", key="session_invalid")
        else:
            # Schema version < 3
            if row["correct_weight"] is not None:
                base_scores = scoring.QuestionScores(
                    row["correct_weight"], row["incorrect_weight"], row["blank_weight"]
                )
            else:
                base_scores = None
        if self.schema_version >= 3:
            self._load_solutions_permutations_scores()
        else:
            # Schema version < 3
            for row in cursor.execute("SELECT * FROM Solutions"):
                self.exam_config.set_solutions(
                    _dec_model(row["model"]), row["solutions"]
                )
            for row in cursor.execute("SELECT * FROM Permutations"):
                self.exam_config.set_permutations(
                    _dec_model(row["model"]), row["permutations"]
                )
            if base_scores is not None:
                # This must be done after having set the solutions
                self.exam_config.set_base_scores(base_scores, same_weights=True)
        if self.schema_version >= 6:
            self._load_model_variations()
        return self.exam_config

    def _load_solutions_permutations_scores(self):
        cursor = self.conn.cursor()
        scores_mode = self.exam_config.scores_mode
        solutions = {}
        permutations = {}
        variations = {}
        scores = {}
        if scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
            weights = {}
        elif scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL:
            scores = {}
        models = []
        model = None
        for row in cursor.execute("SELECT * FROM Questions ORDER BY model, question"):
            if row["model"] != model:
                model = row["model"]
                models.append(model)
                model_permutations = []
                permutations[model] = model_permutations
                if self.schema_version >= 5:
                    model_variations = []
                    variations[model] = model_variations
                if scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
                    model_weights = []
                    weights[model] = model_weights
                elif scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL:
                    model_scores = []
                    scores[model] = model_scores
            model_permutations.append(row["permutation"])
            if self.schema_version >= 5:
                model_variations.append(row["variation"])
            if scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
                model_weights.append(row["score_weight"])
            elif scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL:
                question_scores = scoring.QuestionScores(
                    row["score_correct"], row["score_incorrect"], row["score_blank"]
                )
                model_scores.append(question_scores)
        for m in models:
            model = _dec_model(m)
            solutions = self._load_solutions(m)
            if solutions[0] is not None:
                self.exam_config.set_solutions(model, solutions)
            if permutations[m][0] is not None:
                self.exam_config.set_permutations(model, permutations[m])
            if m in variations:
                self.exam_config.set_question_variations(model, variations[m])
            if scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
                self.exam_config.set_question_weights(model, weights[m])
            elif scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL:
                self.exam_config.set_question_scores(model, scores[m])

    def _load_solutions(self, model):
        solutions = [set() for _ in range(self.exam_config.num_questions)]
        cursor = self.conn.cursor()
        if self.schema_version >= 5:
            query = (
                "SELECT question, solution "
                "FROM Solutions "
                "WHERE model=? ORDER BY question"
            )
        else:
            query = (
                "SELECT question, solution "
                "FROM Questions "
                "WHERE model=? ORDER BY question"
            )
        for row in cursor.execute(query, (model,)):
            solutions[row["question"]].add(row["solution"])
        # Swap empty solutions by None:
        solutions = [s if s else None for s in solutions]
        return solutions

    def _load_model_variations(self):
        cursor = self.conn.cursor()
        variations = {}
        for row in cursor.execute("SELECT * FROM Models"):
            model = _dec_model(row["model"])
            variations[model] = row["model_variation"]
        if variations:
            self.exam_config.model_variations = variations

    def _update_answer(self, exam_id, question, new_answer, commit=True):
        cursor = self.conn.cursor()
        cursor.execute(
            "UPDATE Answers SET answer = ?" "WHERE exam_id = ? AND question = ?",
            (new_answer, exam_id, question),
        )
        if commit:
            self.conn.commit()

    def _update_score(self, exam_id, score, commit=True):
        cursor = self.conn.cursor()
        cursor.execute(
            "UPDATE Exams SET correct = ?, incorrect = ?,"
            "                 blank = ?, score = ?"
            "WHERE exam_id = ?",
            (score.correct, score.incorrect, score.blank, score.score, exam_id),
        )
        if commit:
            self.conn.commit()

    def _group_max_seq(self, group_id):
        cursor = self.conn.cursor()
        cursor.execute(
            "SELECT MAX(sequence_num) FROM Students " "WHERE group_id = ?", (group_id,)
        )
        result = cursor.fetchone()[0]
        if result is not None:
            return int(result)
        else:
            return 0

    def _store_answers(self, exam_id, answers, commit=True):
        data = []
        for i, answer in enumerate(answers):
            data.append((exam_id, i, answer))
        if data:
            cursor = self.conn.cursor()
            cursor.executemany("INSERT INTO Answers VALUES (?, ?, ?)", data)
            if commit:
                self.conn.commit()

    def _store_answer_cells(self, exam_id, answer_cells, commit=True):
        data = []
        for i, question_cells in enumerate(answer_cells):
            for j, cell in enumerate(question_cells):
                item = (
                    exam_id,
                    i,
                    j,
                    cell.center[0],
                    cell.center[1],
                    cell.diagonal,
                    cell.plu[0],
                    cell.plu[1],
                    cell.pru[0],
                    cell.pru[1],
                    cell.pld[0],
                    cell.pld[1],
                    cell.prd[0],
                    cell.prd[1],
                )
                data.append(item)
        if data:
            cursor = self.conn.cursor()
            cursor.executemany(
                "INSERT INTO AnswerCells VALUES "
                "(?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
                data,
            )
            if commit:
                self.conn.commit()

    def _store_id_cells(self, exam_id, id_cells, commit=True):
        if id_cells:
            data = []
            for i, cell in enumerate(id_cells):
                item = (
                    exam_id,
                    i,
                    cell.plu[0],
                    cell.plu[1],
                    cell.pru[0],
                    cell.pru[1],
                    cell.pld[0],
                    cell.pld[1],
                    cell.prd[0],
                    cell.prd[1],
                )
                data.append(item)
            cursor = self.conn.cursor()
            cursor.executemany(
                "INSERT INTO IdCells VALUES " "(?, ?, ?, ?, ?, ?, ?, ?, ?, ?)", data
            )
            if commit:
                self.conn.commit()

    def _enable_foreign_key_constrains(self):
        cursor = self.conn.cursor()
        cursor.execute("PRAGMA foreign_keys=ON")

    def _student_from_row(self, row):
        if self.schema_version >= 2:
            student = students.Student(
                row["student_id"],
                row["full_name"],
                row["first_name"],
                row["last_name"],
                row["email"],
                db_id=row["db_id"],
                group_id=row["group_id"],
                sequence_num=row["sequence_num"],
                is_in_database=True,
            )
        else:
            student = students.Student(
                row["student_id"],
                row["name"],
                None,
                None,
                row["email"],
                db_id=row["db_id"],
                group_id=row["group_id"],
                sequence_num=row["sequence_num"],
                is_in_database=True,
            )
        return student


class ExamFromDB(exams.Exam):
    def __init__(self, db_dict, students_rank, sessiondb):
        """Creates a new ExamFromDB object.

        For efficiency reasons, the 'capture' is not loaded. Use
        'load_capture()' to load it if needed.

        """
        self.sessiondb = sessiondb
        self.capture = None
        self.student_listings = sessiondb.student_listings
        self.exam_id = db_dict["exam_id"]
        self.model = db_dict["model"]
        if db_dict["student_id"]:
            student = sessiondb.student_listings.student(db_dict["student_id"])
        else:
            student = None
        answers = sessiondb.read_answers(self.exam_id)
        self.decisions = ExamDecisionsFromDB(
            answers, student, students_rank, _dec_model(db_dict["model"])
        )
        solutions = sessiondb.exam_config.get_solutions(self.decisions.model)
        if (
            self.decisions.model
            and self.decisions.model in sessiondb.exam_config.scores
        ):
            question_scores = sessiondb.exam_config.scores[self.decisions.model]
        else:
            question_scores = None
        self.score = scoring.Score(answers, solutions, question_scores)


class ExamDecisionsFromDB(capture.ExamDecisions):
    def __init__(self, answers, student, students_rank, model):
        self.answers = answers
        self.student = student
        self.model = model
        self.detected_id = None
        self.id_scores = None
        self.students_rank = students_rank


class GroupListingFromDB(students.GroupListing):
    def __init__(self, sessiondb, group, student_list):
        super().__init__(group, student_list)
        self.sessiondb = sessiondb

    def add_students(self, student_list):
        super().add_students(student_list)
        self.sessiondb.store_students(student_list)

    def rename(self, new_name):
        super().rename(new_name)
        self.sessiondb.rename_student_group(self.group)


class StudentListingsFromDB(students.StudentListings):
    def __init__(self, sessiondb):
        super().__init__()
        self.sessiondb = sessiondb

    def create_listing(self, group):
        created_group = self.sessiondb.create_student_group(group.name)
        group.identifier = created_group.identifier
        listing = GroupListingFromDB(self.sessiondb, group, [])
        self.add_listing(listing)
        return listing

    def remove_at(self, index):
        group = self.listings[index].group
        num_exams = self.sessiondb.num_exams_from_group(group)
        if num_exams == 0:
            super().remove_at(index)
            self.sessiondb.remove_group(group)
        else:
            raise students.CantRemoveGroupException(
                "Some exams ({}) belong to the group".format(num_exams)
            )


def _enc_model(model_letter: str) -> int:
    if model_letter == "0":
        return 0
    elif model_letter is None or model_letter == "?":
        return -1
    else:
        return ord(model_letter) - 64


def _dec_model(model_number: int) -> Optional[str]:
    if model_number == 0:
        return "0"
    elif model_number == -1 or model_number is None:
        return None
    else:
        return chr(64 + model_number)


def check_file_is_sqlite(filename):
    try:
        with open(filename, "rb") as f:
            data = f.read(16)
        if data == b"SQLite format 3\x00":
            is_sqlite = True
        else:
            is_sqlite = False
    except IOError:
        is_sqlite = False
    return is_sqlite


def create_session_directory(dir_name, exam_data, student_listings):
    """Create the session database and directory layout.

    `dir_name` must be an empty directory that already exists. If
    it does not exist, it is created here.

    """
    if not os.path.isdir(dir_name):
        os.mkdir(dir_name)
    os.mkdir(os.path.join(dir_name, "captures"))
    os.mkdir(os.path.join(dir_name, "internal"))
    db_file = os.path.join(dir_name, "session.eyedb")
    _create_session_db(db_file, exam_data, student_listings)


def _create_session_db(db_file, exam_data, student_listings):
    conn = sqlite3.connect(db_file)
    conn.row_factory = sqlite3.Row
    _create_tables(conn)
    _save_exam_config(conn, exam_data)
    _save_student_listings(conn, student_listings)
    conn.commit()
    conn.close()


def _create_tables(conn):
    cursor = conn.cursor()
    cursor.execute(SessionDB._table_session)
    cursor.execute(SessionDB._table_questions)
    cursor.execute(SessionDB._table_solutions)
    cursor.execute(SessionDB._table_models)
    cursor.execute(SessionDB._table_exams)
    cursor.execute(SessionDB._table_students)
    cursor.execute(SessionDB._table_student_groups)
    cursor.execute(SessionDB._table_answers)
    cursor.execute(SessionDB._table_answer_cells)
    cursor.execute(SessionDB._table_id_cells)
    cursor.execute(SessionDB._index_student_id)


def _base_scores(exam_data):
    if exam_data.base_scores is None:
        base_scores = (None, None, None)
    else:
        base_scores = (
            exam_data.base_scores.format_correct_score(),
            exam_data.base_scores.format_incorrect_score(),
            exam_data.base_scores.format_blank_score(),
        )
    return base_scores


def _question_scores(exam_data, model):
    all_none = exam_data.num_questions * [None]
    if exam_data.scores_mode == exams.ExamConfig.SCORES_MODE_INDIVIDUAL:
        weights = all_none
        scores_c = [s.format_correct_score() for s in exam_data.scores[model]]
        scores_i = [s.format_incorrect_score() for s in exam_data.scores[model]]
        scores_b = [s.format_blank_score() for s in exam_data.scores[model]]
    else:
        if exam_data.scores_mode == exams.ExamConfig.SCORES_MODE_WEIGHTS:
            weights = exam_data.get_question_weights(model, formatted=True)
        else:
            weights = all_none
        scores_c = all_none
        scores_i = all_none
        scores_b = all_none
    return scores_c, scores_i, scores_b, weights


def _save_exam_config(conn, exam_data):
    base_scores = _base_scores(exam_data)
    cursor = conn.cursor()
    cursor.execute(
        "INSERT INTO Session "
        "VALUES (?, ?, NULL, NULL, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)",
        (
            SessionDB.DB_SCHEMA_VERSION,
            utils.version,
            exam_data.format_dimensions(),
            SessionDB.GRADING_MODE_ONE_CORRECT,
            exam_data.scores_mode,
            base_scores[0],
            base_scores[1],
            base_scores[2],
            exam_data.id_num_digits,
            1 if exam_data.survey_mode else 0,
            1 if exam_data.left_to_right_numbering else 0,
            exam_data.capture_pattern,
        ),
    )
    # Store the model variations
    if exam_data.model_variations:
        data = []
        for model, variation in exam_data.model_variations.items():
            data.append((_enc_model(model), variation))
        cursor.executemany("INSERT INTO Models VALUES (?, ?)", data)
    # Store the question permutations and scores
    data = []
    for model in exam_data.models:
        all_model = exam_data.num_questions * [_enc_model(model)]
        all_none = exam_data.num_questions * [None]
        permutations = exam_data.get_permutations(model)
        if permutations:
            permutations = [exam_data.format_permutation(p) for p in permutations]
        else:
            permutations = all_none
        question_variations = exam_data.get_question_variations(model)
        scores_c, scores_i, scores_b, weights = _question_scores(exam_data, model)
        data.extend(
            zip(
                all_model,
                range(exam_data.num_questions),
                permutations,
                question_variations,
                scores_c,
                scores_i,
                scores_b,
                weights,
            )
        )
    cursor.executemany("INSERT INTO Questions VALUES (?, ?, ?, ?, ?, ?, ?, ?)", data)
    # Store question solutions
    data = []
    for model in exam_data.models:
        encoded_model = _enc_model(model)
        solutions = exam_data.get_solutions(model)
        if solutions:
            for question, question_solutions in enumerate(solutions):
                for solution in question_solutions:
                    data.append((encoded_model, question, solution))
    cursor.executemany("INSERT INTO Solutions VALUES (?, ?, ?)", data)


def _save_student_listings(conn, student_listings):
    for group_listing in student_listings:
        _save_group_listing(conn, group_listing)


def _save_group_listing(conn, group_listing):
    cursor = conn.cursor()
    cursor.execute(
        "INSERT INTO StudentGroups (group_id, group_name) VALUES (?, ?)",
        (group_listing.group.identifier, group_listing.group.name),
    )
    if len(group_listing) > 0:
        internal_list = []
        for student in group_listing.students:
            internal_list.append(
                (
                    student.student_id,
                    student.full_name,
                    student.first_name,
                    student.last_name,
                    student.email,
                    student.group_id,
                    student.sequence_num,
                )
            )
        cursor.executemany(
            "INSERT INTO Students "
            "(student_id, full_name, "
            " first_name, last_name, email, group_id, "
            " sequence_num) VALUES "
            "(?, ?, ?, ?, ?, ?, ?)",
            internal_list,
        )


def _create_cell_from_row(row, is_id_cell=False):
    plu = (row["lux"], row["luy"])
    pru = (row["rux"], row["ruy"])
    pld = (row["ldx"], row["ldy"])
    prd = (row["rdx"], row["rdy"])
    if not is_id_cell:
        center = (row["center_x"], row["center_y"])
        diagonal = row["diagonal"]
    else:
        center = None
        diagonal = None
    return capture.CellGeometry(plu, pru, pld, prd, center, diagonal)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2024 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import unittest

import eyegrade.stats.stats_core as stats_core


class TestQuestionStats(unittest.TestCase):
    def testQuestionStatsWithPermutations(self):
        permutations = stats_core.ChoicePermutations(
            num_choices=4,
            reference_model="0",
            choice_permutations={
                "A": [1, 4, 2, 3],
                "B": [3, 2, 4, 1],
            },
        )
        answers = [
            ("A", 1),
            ("B", 3),
            ("A", 0),
            ("B", 3),
            ("B", 2),
            ("A", 2),
        ]
        correct_choices = {
            "0": [4],
        }
        stats = stats_core.QuestionStats(4, correct_choices, permutations)
        for model, answer in answers:
            stats.count_answer(answer, model)
        self.assertEqual(stats.get_answer_counts("0"), [1, 1, 1, 0, 3])
        self.assertEqual(stats.get_answer_counts("A"), [1, 1, 3, 1, 0])
        self.assertEqual(stats.get_answer_counts("B"), [1, 0, 1, 3, 1])
        self.assertTrue(stats.is_correct(4, "0"))
        self.assertTrue(stats.is_correct(2, "A"))
        self.assertTrue(stats.is_correct(3, "B"))
        self.assertFalse(stats.is_correct(2, "0"))
        self.assertFalse(stats.is_correct(1, "A"))
        self.assertFalse(stats.is_correct(4, "B"))

    def testQuestionStatsWithoutPermutations(self):
        answers = [
            ("A", 1),
            ("B", 3),
            ("A", 0),
            ("B", 3),
            ("B", 2),
            ("A", 2),
        ]
        correct_choices = {
            "A": [3],
            "B": [2],
        }
        stats = stats_core.QuestionStats(4, correct_choices, None)
        for model, answer in answers:
            stats.count_answer(answer, model)
        self.assertEqual(stats.get_answer_counts("A"), [1, 1, 1, 0, 0])
        self.assertEqual(stats.get_answer_counts("B"), [0, 0, 1, 2, 0])
        self.assertTrue(stats.is_correct(3, "A"))
        self.assertTrue(stats.is_correct(2, "B"))
        self.assertFalse(stats.is_correct(1, "A"))
        self.assertFalse(stats.is_correct(3, "B"))


class TestExamStats(unittest.TestCase):
    def testExamStatsWithPermutations(self):
        permutations = stats_core.ExamPermutations(
            num_questions=4,
            num_choices=3,
            reference_model="0",
            exam_permutations={
                "A": [
                    (1, [1, 2, 3]),
                    (4, [3, 2, 1]),
                    (3, [2, 1, 3]),
                    (2, [3, 1, 2]),
                ],
                "B": [
                    (2, [1, 3, 2]),
                    (3, [2, 3, 1]),
                    (4, [1, 2, 3]),
                    (1, [3, 2, 1]),
                ],
            },
        )
        answers = [
            ("A", [1, 3, 2, 0]),
            ("B", [1, 2, 1, 3]),
            ("A", [2, 2, 2, 1]),
            ("B", [3, 0, 1, 1]),
            ("B", [1, 2, 1, 1]),
            ("A", [1, 3, 2, 3]),
        ]
        correct_choices = {
            "0": [[1], [3], [2], [3]],
        }
        stats = stats_core.QByQStats(4, 3, correct_choices, permutations)
        for model, answers in answers:
            stats.count_answers(answers, model)
        self.assertEqual(
            stats.get_answer_counts("0"),
            [
                [0, 3, 1, 2],
                [1, 2, 2, 1],
                [1, 3, 0, 2],
                [0, 5, 1, 0],
            ],
        )
        self.assertEqual(
            stats.get_answer_counts("A"),
            [
                [0, 3, 1, 2],
                [0, 0, 1, 5],
                [1, 0, 3, 2],
                [1, 1, 2, 2],
            ],
        )
        self.assertTrue(stats.is_correct(1, 1, "0"))
        self.assertTrue(stats.is_correct(3, 2, "0"))
        self.assertTrue(stats.is_correct(2, 3, "0"))
        self.assertTrue(stats.is_correct(3, 4, "0"))
        self.assertFalse(stats.is_correct(3, 1, "0"))
        self.assertFalse(stats.is_correct(2, 2, "0"))
        self.assertFalse(stats.is_correct(1, 3, "0"))
        self.assertFalse(stats.is_correct(1, 4, "0"))
        self.assertTrue(stats.is_correct(1, 2, "A"))
        self.assertTrue(stats.is_correct(3, 4, "B"))
        self.assertFalse(stats.is_correct(2, 2, "A"))
        self.assertFalse(stats.is_correct(1, 4, "B"))

    def testExamStatsWithoutPermutations(self):
        answers = [
            ("A", [1, 3, 2, 0]),
            ("B", [1, 2, 1, 3]),
            ("A", [2, 2, 2, 1]),
            ("B", [3, 0, 1, 1]),
            ("B", [1, 2, 1, 1]),
            ("A", [1, 3, 2, 3]),
        ]
        correct_choices = {
            "A": [[1], [3], [2], [3]],
            "B": [[2], [1], [1], [3]],
        }
        stats = stats_core.QByQStats(4, 3, correct_choices, None)
        for model, answers in answers:
            stats.count_answers(answers, model)
        self.assertEqual(
            stats.get_answer_counts("A"),
            [
                [0, 2, 1, 0],
                [0, 0, 1, 2],
                [0, 0, 3, 0],
                [1, 1, 0, 1],
            ],
        )
        self.assertEqual(
            stats.get_answer_counts("B"),
            [
                [0, 2, 0, 1],
                [1, 0, 2, 0],
                [0, 3, 0, 0],
                [0, 2, 0, 1],
            ],
        )

    def testExamStatsNoExams(self):
        permutations = stats_core.ExamPermutations(
            num_questions=4,
            num_choices=3,
            reference_model="0",
            exam_permutations={
                "A": [
                    (1, [1, 2, 3]),
                    (4, [3, 2, 1]),
                    (3, [2, 1, 3]),
                    (2, [3, 1, 2]),
                ],
                "B": [
                    (2, [1, 3, 2]),
                    (3, [2, 3, 1]),
                    (4, [1, 2, 3]),
                    (1, [3, 2, 1]),
                ],
            },
        )
        correct_choices = {
            "0": [[1], [3], [2], [3]],
        }
        stats = stats_core.QByQStats(4, 3, correct_choices, permutations)
        self.assertEqual(
            stats.get_answer_counts("0"),
            [
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ],
        )
        self.assertEqual(
            stats.get_answer_ratios("0"),
            [
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
                [0.0, 0.0, 0.0, 0.0],
            ],
        )
        self.assertEqual(
            stats.get_answer_counts("A"),
            [
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
                [0, 0, 0, 0],
            ],
        )



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import math
import copy
import sys
import itertools

import cv2
import numpy as np

# Local imports
from . import geometry as g
from . import capture
from . import sessiondb
from . import images
from . import utils
from .ocr import classifiers
from .ocr import sample

# Adaptive threshold algorithm
param_adaptive_threshold_block_size = 45
param_adaptive_threshold_offset = 4

# Other detection parameters
param_collapse_lines_maxgap = 7
param_directions_threshold = 0.4
param_hough_thresholds = [280, 260, 240, 225, 210, 195, 180, 160, 140, 120]
param_failures_threshold = 10
param_check_corners_tolerance_mul = 6

# Parameters for the infobits masks
param_bit_mask_threshold = 0.25
param_bit_mask_radius_multiplier = 0.333

# Parameters for id boxes detection
param_id_boxes_min_energy_threshold = 0.5
param_id_boxes_mean_energy_threshold = 0.75
param_id_boxes_energy_break = 0.99
param_id_boxes_min_height = 15
param_id_boxes_discard_distance = 20

# Other parameters
param_error_log = "eyegrade-errors.log"
param_error_image_pattern = "error-%s.png"


class ExamDetector:
    default_options = {
        "infobits": True,
        "left-to-right-numbering": False,
        "show-lines": False,
        "debug-ocr": False,
        "show-image-proc": False,
        "read-id": False,
        "show-status": False,
        "capture-from-file": False,
        "capture-raw-file": None,
        "capture-proc-file": None,
        "capture-proc-ipl": None,
        "error-logging": False,
        "logging-dir": ".",
    }

    @classmethod
    def get_default_options(cls):
        return copy.copy(cls.default_options)

    def __init__(self, dimensions, context, options, image_raw=None):
        self.options = options
        self.context = context
        if image_raw is not None:
            self.image_raw = image_raw
            self.image_proc = pre_process(self.image_raw)
        elif not self.options["capture-from-file"]:
            self.image_raw = self.context.capture()
            self.image_proc = pre_process(self.image_raw)
        elif self.options["capture-raw-file"] is not None:
            self.image_raw = images.load_image(self.options["capture-raw-file"])
            if self.image_raw is None:
                raise utils.EyegradeException("", key="load_image")
            self.image_proc = pre_process(self.image_raw)
        elif self.options["capture-proc-file"] is not None:
            self.image_raw = images.load_image(self.options["capture-proc-file"])
            self.image_proc = images.rgb_to_gray(self.image_raw)
        elif self.options["capture-proc-ipl"] is not None:
            self.image_raw = self.options["capture-proc-ipl"]
            self.image_proc = self.options["capture-proc-ipl"]
        else:
            raise Exception("Wrong capture options")
        self.dimensions = dimensions
        self.status = {
            "lines": False,
            "boxes": False,
            "cells": False,
            "infobits": False,
            "id-box-hlines": False,
            "id-box": False,
        }
        if self.options["show-image-proc"]:
            self.image_to_show = images.gray_to_rgb(self.image_proc)
        elif self.options["show-lines"]:
            self.image_to_show = self.image_raw.copy()
        else:
            self.image_to_show = self.image_raw
        self.decisions = None
        self.capture = None

    def detect_safe(self):
        try:
            return self.detect()
        except Exception:
            self.success = False
            self.status["cells"] = False
            self.status["infobits"] = False
            self.context.notify_failure()
            if self.options["error-logging"]:
                exc_type, exc_value, exc_traceback = sys.exc_info()
                self._write_error_trace(exc_type, exc_value, exc_traceback)
            # else... silence the exception, and try with the next capture

    def detect(self):
        answers = None
        detected_id = None
        id_scores = None
        bits = None
        answer_cells = None
        id_cells = None
        id_hlines = None
        success = False
        axes = None
        lines = detect_lines(self.image_proc, self.context.get_hough_threshold())
        if len(lines) >= 2:
            self.status["lines"] = True
            axes = detect_boxes(lines, self.dimensions)
        if axes is None:
            self.context.next_hough_threshold()
        else:
            self.status["boxes"] = True
            axes = filter_axes(
                axes,
                images.get_width(self.image_raw),
                images.get_height(self.image_raw),
                self.options["read-id"],
            )
            corner_matrixes = cell_corners(
                axes[1][1],
                axes[0][1],
                images.get_width(self.image_raw),
                images.get_height(self.image_raw),
                self.dimensions,
            )
            if len(corner_matrixes) > 0:
                self.status["cells"] = True
                answer_cells = self._answer_cells_geometry(corner_matrixes)
                answers = self._decide_cells(answer_cells)
                if self.options["infobits"]:
                    bits = read_infobits(self.image_proc, corner_matrixes)
                    if bits is not None:
                        self.status["infobits"] = True
                        success = True
                    else:
                        success = False
                else:
                    success = True
                if success and self.options["read-id"]:
                    id_hlines, id_cells = id_boxes_geometry(
                        self.image_proc,
                        self.options["id-num-digits"],
                        axes[1][1],
                        self.dimensions,
                    )
                    if id_hlines:
                        self.status["id-box-hlines"] = True
                    if not id_cells:
                        success = False
                    else:
                        self.status["id-box"] = True
                        detected_id, id_scores = self._detect_id(id_cells)
                else:
                    id_cells = []
        if success:
            self.context.notify_success()
        else:
            self.context.notify_failure()
        # Draw debug information on the capture
        if self.options["show-lines"]:
            if self.status["cells"]:
                for line in axes[0][1]:
                    images.draw_line(self.image_to_show, line, (255, 0, 0))
                for line in axes[1][1]:
                    images.draw_line(self.image_to_show, line, (255, 0, 255))
                self._draw_cell_corners(corner_matrixes)
            elif self.status["lines"]:
                for line in lines:
                    images.draw_line(self.image_to_show, line, (255, 0, 0))
            if id_hlines:
                for line in id_hlines:
                    images.draw_line(self.image_to_show, line, (255, 255, 0))
            if id_cells:
                for cell in id_cells:
                    for corner in cell.corners():
                        images.draw_point(self.image_to_show, corner)
        if self.options["show-status"]:
            self._draw_status_flags()
        self.decisions = capture.ExamDecisions(
            success, answers, detected_id, id_scores, infobits=bits
        )
        self.capture = capture.ExamCapture(
            self.image_to_show, answer_cells, id_cells, self._compute_progress()
        )
        self.success = success
        return success

    def detect_manual(self, manual_points):
        """Called when cell corners are obtained from manual detection."""
        bits = None
        success = False
        answers = None
        answer_cells = None
        corner_matrixes = process_box_corners(manual_points, self.dimensions)
        if corner_matrixes != []:
            self.status["cells"] = True
            answer_cells = self._answer_cells_geometry(corner_matrixes)
            answers = self._decide_cells(answer_cells)
            if self.options["infobits"]:
                bits = read_infobits(self.image_proc, corner_matrixes)
                if bits is not None:
                    self.status["infobits"] = True
                    success = True
            else:
                success = True
        id_cells = None
        detected_id = None
        id_scores = None
        self.decisions = capture.ExamDecisions(
            success, answers, detected_id, id_scores, infobits=bits
        )
        self.capture = capture.ExamCapture(
            self.image_to_show, answer_cells, id_cells, 1.0
        )
        self.success = success
        return success

    def try_to_detect(self):
        """Checks if the image has a probable exam.

        Sets `self.exam_detected` to True if an exam is detected as
        probable. Right now, it just checks that the proper axes are
        detected.

        """
        self.exam_detected = False
        lines = detect_lines(self.image_proc, self.context.get_hough_threshold())
        if len(lines) >= 2:
            axes = detect_boxes(lines, self.dimensions)
            if axes is not None:
                self.exam_detected = True

    def _write_error_trace(self, exc_type, exc_value, exc_traceback):
        import datetime
        import re
        import traceback
        import os

        if not self.options["capture-from-file"]:
            print("Exception catched! Storing trace into a log file...")
            date = str(datetime.datetime.now())
            logname = os.path.join(self.options["logging-dir"], param_error_log)
            file_ = open(logname, "a")
            file_.write("-" * 60 + "\n")
            file_.write(date + "\n")
            file_.write("Hough threshold: %d\n" % self.context.get_hough_threshold())
            traceback.print_exception(exc_type, exc_value, exc_traceback, file=file_)
            file_.close()
            im_file = param_error_image_pattern % re.sub(r"[-\ \.:]", "_", date)
            capture.save_image(
                os.path.join(self.options["logging-dir"], im_file), self.image_raw
            )
        else:
            traceback.print_exception(exc_type, exc_value, exc_traceback)

    def _answer_cells_geometry(self, corner_matrixes):
        cells = []
        for corners in corner_matrixes:
            for i in range(0, len(corners) - 1):
                row = []
                for j in range(0, len(corners[0]) - 1):
                    cell = capture.CellGeometry(
                        corners[i][j],
                        corners[i][j + 1],
                        corners[i + 1][j],
                        corners[i + 1][j + 1],
                        None,
                        None,
                    )
                    row.append(cell)
                cells.append(row)
        if self.options["left-to-right-numbering"]:
            cells = self._set_left_to_right(cells)
        return cells

    def _decide_cells(self, answer_cells):
        decisions = []
        for row in answer_cells:
            row_decisions = []
            for cell in row:
                corners = np.array([cell.plu, cell.pru, cell.pld, cell.prd])
                samp = sample.CrossSampleFromCam(corners, self.image_proc)
                decision = self.context.crosses_classifier.is_cross(samp)
                row_decisions.append(decision)
            decisions.append(decide_answer(row_decisions))
        return decisions

    def _set_left_to_right(self, cells):
        """Sets left to right order in cell geometry."""
        cells_transposed = []
        num_rows = max([questions for choices, questions in self.dimensions])
        heads = [1]
        for choices, questions in self.dimensions:
            heads.append(heads[-1] + questions)
        heads.append(len(cells) + 1)
        for row in range(0, num_rows):
            for column in range(0, len(self.dimensions)):
                pos = heads[column] + row
                if pos < heads[column + 1]:
                    cells_transposed.append(cells[pos - 1])
        return cells_transposed

    def _compute_progress(self):
        progress = 0
        if self.status["lines"]:
            progress += 1
        if self.status["boxes"]:
            progress += 1
        if self.status["cells"]:
            progress += 1
        if self.status["infobits"]:
            progress += 1
        if self.status["id-box-hlines"]:
            progress += 1
        if self.status["id-box"]:
            progress += 1
        max_progress = 6
        if not self.options["read-id"]:
            max_progress -= 2
        if not self.options["infobits"]:
            max_progress -= 1
        return progress / max_progress

    def _detect_id(self, id_cells):
        if id_cells is None:
            detected_id = None
        digits = []
        id_scores = []
        for cell in id_cells:
            corners = np.array([cell.plu, cell.pru, cell.pld, cell.prd])
            samp = sample.DigitSampleFromCam(corners, self.image_proc)
            digit, scores = self.context.ocr.classify_digit(samp)
            digits.append(digit)
            id_scores.append(scores)
        detected_id = "".join([str(d) if d is not None else "0" for d in digits])
        return detected_id, id_scores

    def _draw_status_flags(self):
        flags = []
        flags.append(("L", self.status["lines"]))
        flags.append(("B", self.status["boxes"]))
        flags.append(("C", self.status["cells"]))
        flags.append(("M", self.status["infobits"]))
        flags.append(("H", self.status["id-box-hlines"]))
        flags.append(("N", self.status["id-box"]))
        color_good = (255, 0, 0)
        color_bad = (0, 0, 255)
        y = 75
        width = 24
        x = images.get_width(self.image_to_show) - 5 - len(flags) * width
        for letter, value in flags:
            color = color_good if value else color_bad
            images.draw_text(self.image_to_show, letter, color, (x, y))
            x += width

    def _draw_hough_threshold(self):
        pos = (images.get_width(self.image_to_show) - 77, 110)
        images.draw_text(
            self.image_to_show, str(self.context.get_hough_threshold()), position=pos
        )

    def _draw_cell_corners(self, corner_matrixes):
        for corners in corner_matrixes:
            for h in corners:
                for c in h:
                    images.draw_point(self.image_to_show, c)


class ImageTransformer:
    """Apply transformations to the image captured by the webcam.

    Available transformations are: flipping around the x axis, the y axis or both.

    By default, no transformation gets applied.

    """

    IDENTITY = 2
    FLIP_V = 0
    FLIP_H = 1
    FLIP_BOTH = -1

    def __init__(self, transformation):
        if transformation < -1 or transformation > 2:
            raise ValueError("Unknown transformation: {}".format(transformation))
        self.transformation = transformation

    def transform(self, image):
        """Apply the configured transformation to an image and return a new image.

        In case of the identity transformation, a reference to the same image is returned
        (it isn't copied into a new image).

        """
        if self.transformation == ImageTransformer.IDENTITY:
            dst_image = image
        else:
            dst_image = cv2.flip(image, self.transformation)
        return dst_image


class ExamDetectorContext:
    """Class intended for persistency of data accross several
    ExamCapture objects.

    """

    def __init__(
        self,
        camera_id=-1,
        fixed_hough_threshold=None,
        image_transformer=ImageTransformer(ImageTransformer.IDENTITY),
    ):
        """Creates a new camera capture context.

        A default initial camera can be specified with `camera_id` (an
        integer). Pass -1 (the default value) for letting this object
        choose the first available camera.

        """
        if not fixed_hough_threshold:
            self.hough_thresholds = param_hough_thresholds
        else:
            self.hough_thresholds = [fixed_hough_threshold]
        self.hough_thresholds_idx = 0
        self.failures_in_a_row = 0
        self.camera = None
        self.camera_id = camera_id
        self.threshold_locked = False
        self.image_transformer = image_transformer
        self.ocr = classifiers.DefaultDigitClassifier()
        self.crosses_classifier = classifiers.DefaultCrossesClassifier()

    def open_camera(self, camera_id=None):
        """Initializes the last camera device used, or `camera_id`.

        If no camera has been used in the past, the initial camera id
        received in the constructor is used.  Returns True if there is
        success. If a camera is already open, it does nothing and
        returns True. If the camera does not work anymore, it tries
        with other cameras.

        """
        previous_camera = self.camera_id
        if camera_id is not None:
            self.camera_id = camera_id
            if self.camera is not None:
                self.close_camera()
        if self.camera is None:
            if self.camera_id is not None and self.camera_id != -1:
                self.camera = self._try_camera(self.camera_id)
            if self.camera is None:
                if previous_camera is not None and previous_camera >= 0:
                    self.camera_id = previous_camera
                    self.camera = self._try_camera(previous_camera)
                    if self.camera is None:
                        self.camera, self.camera_id = self._try_next_camera(-1)
        return self.camera is not None

    def current_camera_id(self):
        """Returns the id (integer) of the current camera or None."""
        if self.camera_id >= 0:
            return self.camera_id
        else:
            return None

    def next_camera(self):
        """Selects the next camera available.

        Note that changing camera does not seem to work currently
        in OpenCV, at least in Linux.

        """

        if self.camera is not None:
            del self.camera
        camera, camera_id = self._try_next_camera(self.camera_id)
        if camera is not None:
            self.camera, self.camera_id = camera, camera_id
            return True
        else:
            return False

    def apply_image_transfomer(self, image_transfomer):
        """Use the given image tranformer from now on."""
        self.image_transformer = image_transfomer

    def lock_threshold(self):
        self.threshold_locked = True

    def unlock_threshold(self):
        self.threshold_locked = False
        self.failures_in_a_row = 0

    def get_hough_threshold(self):
        return self.hough_thresholds[self.hough_thresholds_idx]

    def next_hough_threshold(self):
        if not self.threshold_locked:
            self.hough_thresholds_idx = (self.hough_thresholds_idx + 1) % len(
                self.hough_thresholds
            )
            self.failures_in_a_row = 0

    def notify_failure(self):
        self.failures_in_a_row += 1
        if self.failures_in_a_row > param_failures_threshold:
            self.next_hough_threshold()

    def notify_success(self):
        self.failures_in_a_row = 0

    def close_camera(self):
        """Closes the current camera.

        The same camera will be opened again when open_camera() is called.

        """
        if self.camera is not None:
            self.camera.release()
        self.camera = None

    def capture(self, clone=False, resize=None):
        """Returns a capture.

        If `clone` is True, the image returned is a copy of the
        original.  Use this option if you plan to modify the image or
        store it for later, because the buffer of the original image
        will be reused by OpenCV por the next capture.

        `resize` is a tuple (width, height). If it is not None, then
        the image is scaled to that size. Scaling implies a new copy
        regardless the value of `clone`.

        """
        image = None
        if resize is not None:
            # The image will be cloned when resizing
            clone = False
        if self.camera is not None:
            image = self.capture_image(clone=clone)
            if image is None:
                image = np.zeros((480, 640, 3), dtype=np.uint8)
            if resize is not None:
                image = cv2.resize(image, resize, interpolation=cv2.INTER_AREA)
        return self.image_transformer.transform(image)

    def dump_buffer(self, delay_suffered):
        if self.camera is not None and delay_suffered > 0.1:
            frames_to_drop = min(8, int(1 + (delay_suffered - 0.1) / 0.04))
            for i in range(0, frames_to_drop):
                self.capture_image(False)

    def capture_image(self, clone=False):
        success, image = self.camera.read()
        if not success:
            image = None
        elif clone:
            return image.copy()
        else:
            return image

    def _try_next_camera(self, cur_camera_id):
        camera = None
        camera_id = -1
        for i in itertools.chain(
            range(cur_camera_id + 1, 10), range(0, cur_camera_id + 1)
        ):
            camera = self._try_camera(i)
            if camera is not None:
                camera_id = i
                break
        return (camera, camera_id)

    @staticmethod
    def _try_camera(camera_id):
        cam = cv2.VideoCapture(camera_id)
        if cam is not None:
            success, image = cam.read()
            if success is None or image is None:
                cam = None
        return cam


class FalseExamDetectorContext(ExamDetectorContext):
    def __init__(self, session_file):
        super().__init__()
        self.session = sessiondb.SessionDB(session_file)
        self.camera_id = 99
        self.camera = "Something"
        self.exams = []
        self.next_exam_idx = 0

    def open_camera(self, camera_id=None):
        self.exams = self.session.read_exams()
        self.next_exam_idx = 0

    def current_camera_id(self):
        return 99

    def next_camera(self):
        return True

    def close_camera(self):
        self.exams = []

    def capture(self, clone=False, resize=None):
        if self.exams:
            exam = self.exams[self.next_exam_idx]
            image = self.session.load_raw_capture(exam.exam_id)
        else:
            image = None
        return image

    def dump_buffer(self, delay_suffered):
        pass

    def notify_success(self):
        super().notify_success()
        self.next_exam_idx += 1
        if self.next_exam_idx == len(self.exams):
            self.next_exam_idx = 0


def pre_process(image):
    gray = images.rgb_to_gray(image)
    thr = cv2.adaptiveThreshold(
        gray,
        255,
        cv2.ADAPTIVE_THRESH_GAUSSIAN_C,
        cv2.THRESH_BINARY_INV,
        param_adaptive_threshold_block_size,
        param_adaptive_threshold_offset,
    )
    return thr


def detect_lines(image, hough_threshold):
    raw_lines = cv2.HoughLines(image, 1, 0.01, hough_threshold)
    if raw_lines is None:
        return []
    lines = raw_lines
    if len(lines[0]) > 500:
        return []
    lines = [(float(l[0][0]), float(l[0][1])) for l in lines]
    return sorted(lines, key=lambda x: x[1])


def detect_directions(lines):
    """Group lines into axes.

    Parameters:
    - lines: a list of tuples (rho, theta) sorted from smaller to bigger
        value of rho.

    Returns the list of detected axes where each ax is a tuple of
    (theta, list of lines), being theta the average angle
    of the lines of the ax. Axes are sorted by angle from horizontal
    to vertical. Lines within each ax are sorted by theta.

    """
    assert len(lines) >= 2
    axes = []
    rho, theta = lines[0]
    axes.append((theta, [(rho, theta)]))
    for rho, theta in lines[1:]:
        if abs(theta - axes[-1][0]) < param_directions_threshold:
            axes[-1][1].append((rho, theta))
        else:
            axes.append((theta, [(rho, theta)]))
    if abs(axes[0][0] - axes[-1][0] + math.pi) < param_directions_threshold:
        axes[0][1].extend([(-rho, theta - math.pi) for rho, theta in axes[-1][1]])
        del axes[-1]
    for i in range(0, len(axes)):
        avg = sum([theta for rho, theta in axes[i][1]]) / len(axes[i][1])
        axes[i] = (avg, sorted(axes[i][1], key=lambda x: abs(x[0])))
    if abs(axes[-1][0] - math.pi) < abs(axes[0][0]):
        axes = axes[-1:] + axes[0:-1]
    return axes


def detect_boxes(lines, dimensions):
    """Classify lines in two groups: horizontal and vertical lines.

    Parameters:
    - lines: a list of tuples (rho, theta) sorted from smaller to bigger
        value of rho.
    - dimensions: a list of boxes dimensions, where each box is a tuple
        (number-of-choices, number of questions). For example:
        [(3, 10), (3, 9)] means that the left-most box has 10 questions
        with 3 options each, and the other box has 9 questions, also with
        3 options.

    Returns None if detection failed or the list of two axes
    [vertical, horizontal] where each ax is a tuple of
    (theta, list of lines), being theta the average angle
    of the lines of the ax. Lines within each ax are sorted by theta.

    """
    v_expected = len(dimensions) + sum([box[0] for box in dimensions])
    h_expected = 1 + max([box[1] for box in dimensions])
    axes = detect_directions(lines)
    axes = [axis for axis in axes if len(axis[1]) >= min(v_expected, h_expected)]
    # If there are spurious axes, try to filter them out:
    if len(axes) == 3:
        # Take the (almost) perpendicular angles
        if g.angles_perpendicular(axes[0][0], axes[1][0]):
            del axes[2]
        elif g.angles_perpendicular(axes[0][0], axes[2][0]):
            del axes[1]
        else:
            del axes[0]
    elif len(axes) == 4:
        # Take the angles which are the closest to 0 and pi/2.
        # The image is assumed to have its horizontal and vertical lines closer
        # to zero or pi/2 than spureous lines.
        axes.sort(key=lambda x: g.distance_closest_axis(x[0], (0, math.pi / 2)))
        axes = axes[:2]
    if len(axes) == 2:
        if g.angles_perpendicular(axes[0][0], axes[1][0]):
            return axes
    else:
        return None


def filter_axes(axes, image_width, image_height, read_id):
    """Filters out lines near borders and lines too close to other lines.

    - axes: [(vlines_angle, vlines), (hlines_angle, hlines)]
    - image_width, image_height: image size
    - read_id: True if the id must be read

    Returns a new axes object with updated lines if success or
    the lines without collapsing if not.

    """
    # First, filter out lines too close to image borders
    axes = (
        (
            axes[0][0],
            [
                l
                for l in axes[0][1]
                if (
                    (abs(l[0]) < 0.97 * image_width and abs(l[0]) > 0.03 * image_width)
                    or not g.angles_perpendicular(math.pi / 2, l[1])
                )
            ],
        ),
        (
            axes[1][0],
            [
                l
                for l in axes[1][1]
                if (
                    (
                        abs(l[0]) < 0.97 * image_height
                        and abs(l[0]) > 0.03 * image_height
                    )
                    or not g.angles_perpendicular(0.0, l[1])
                )
            ],
        ),
    )
    # Now, colapse lines that are too close
    hlines = collapse_lines_angles(axes[1][1], True)
    if hlines is None:
        return axes
    vlines = collapse_lines_angles(axes[0][1], False)
    if vlines is None:
        return axes
    return [(axes[0][0], vlines), (axes[1][0], hlines)]


def collapse_lines_angles(lines, horizontal):
    """Collapses lines that are close together.

    Receives the list of pairs of lines (rho, theta),
    and whether lines are horizontal or not.

    Returns the lines or None if the expected number of
    lines is not matched.

    """
    if len(lines) < 2:
        return None
    main_lines = []
    sum_rho = lines[0][0]
    sum_theta = lines[0][1]
    num_lines = 1
    last_line = lines[0]
    for line in lines[1:]:
        if (
            (
                (horizontal and line[1] > last_line[1])
                or (not horizontal and line[1] < last_line[1])
            )
            and abs(line[0] - last_line[0]) >= 5
        ) or abs(line[0] - last_line[0]) > param_collapse_lines_maxgap:
            main_lines.append((sum_rho / num_lines, sum_theta / num_lines))
            sum_rho = line[0]
            sum_theta = line[1]
            num_lines = 1
        else:
            sum_rho += line[0]
            sum_theta += line[1]
            num_lines += 1
        last_line = line
    main_lines.append((sum_rho / num_lines, sum_theta / num_lines))
    return main_lines


def cell_corners(hlines, vlines, iwidth, iheight, dimensions):
    h_expected = 1 + max([box[1] for box in dimensions])
    v_expected = len(dimensions) + sum([box[0] for box in dimensions])
    if len(vlines) != v_expected:
        if len(vlines) > v_expected and len(vlines) <= v_expected + 2:
            # Remove one or two spurious lines
            vlines = g.discard_spurious_lines(vlines, v_expected)
        else:
            return []
    if len(hlines) < h_expected:
        return []
    elif len(hlines) > h_expected:
        hlines = hlines[-h_expected:]
    corner_matrixes = []
    vini = 0
    for width, height in dimensions:
        corners = []
        for i in range(0, height + 1):
            cpart = []
            corners.append(cpart)
            for j in range(vini, vini + width + 1):
                c = g.intersection(hlines[i], vlines[j])
                cpart.append(c)
        corner_matrixes.append(corners)
        vini += 1 + width
    if check_corners(corner_matrixes, iwidth, iheight):
        return corner_matrixes
    else:
        return []


def check_corners(corner_matrixes, width, height):
    # Check differences between horizontal lines:
    corners = corner_matrixes[(len(corner_matrixes) - 1) // 2]
    ypoints = [row[-1][1] for row in corners]
    difs = []
    difs2 = []
    for i in range(1, len(ypoints)):
        difs.append(ypoints[i] - ypoints[i - 1])
    for i in range(1, len(difs)):
        difs2.append(difs[i] - difs[i - 1])
    max_difs2 = (
        1 + float(max(difs) - min(difs)) / len(difs) * param_check_corners_tolerance_mul
    )
    if max(difs2) > max_difs2:
        return False
    if 0.5 * max(difs) > min(difs):
        return False
    # Check that no points are negative
    for corners in corner_matrixes:
        for row in corners:
            for point in row:
                if (
                    point[0] < 0
                    or point[0] >= width
                    or point[1] < 0
                    or point[1] >= height
                ):
                    return False

    # Check that the sequence of points is coherent
    for corners in corner_matrixes:
        for i in range(0, len(corners) - 1):
            for j in range(0, len(corners[0]) - 1):
                if (
                    corners[i][j][1] >= corners[i + 1][j][1]
                    or corners[i][j + 1][1] >= corners[i + 1][j + 1][1]
                    or corners[i][j][0] >= corners[i][j + 1][0]
                    or corners[i + 1][j][0] >= corners[i + 1][j + 1][0]
                ):
                    return False

    # Success if control reaches here
    return True


def read_infobits(image, corner_matrixes):
    mask = images.new_image(images.get_width(image), images.get_height(image), 1)
    bits = []
    for corners in corner_matrixes:
        for i in range(1, len(corners[0])):
            dx = g.diff_points(corners[-1][i - 1], corners[-1][i])
            dy = g.diff_points(corners[-1][i], corners[-2][i])
            center = g.round_point(
                (
                    corners[-1][i][0] + dx[0] / 2 + dy[0] / 2.6,
                    corners[-1][i][1] + dx[1] / 2 + dy[1] / 2.6,
                )
            )
            bits.append(decide_infobit(image, mask, center, dy))
    # Check validity
    if min([b[0] ^ b[1] for b in bits]) is True:
        return [b[0] for b in bits]
    else:
        return None


def decide_infobit(image, mask, center_up, dy):
    center_down = g.add_points(center_up, dy)
    radius = int(
        round(
            math.sqrt(dy[0] * dy[0] + dy[1] * dy[1]) * param_bit_mask_radius_multiplier
        )
    )
    if radius == 0:
        radius = 1
    images.zero_image(mask)
    cv2.circle(mask, center_up, radius, (1), thickness=-1)
    mask_pixels = cv2.countNonZero(mask)
    masked = cv2.multiply(image, mask)
    masked_pixels_up = cv2.countNonZero(masked)
    images.zero_image(mask)
    cv2.circle(mask, center_down, radius, (1), thickness=-1)
    masked = cv2.multiply(image, mask)
    masked_pixels_down = cv2.countNonZero(masked)
    if mask_pixels < 1:
        return (False, False)
    return (
        float(masked_pixels_up) / mask_pixels >= param_bit_mask_threshold,
        float(masked_pixels_down) / mask_pixels >= param_bit_mask_threshold,
    )


def decide_answer(cell_decisions):
    marked = [i for i in range(0, len(cell_decisions)) if cell_decisions[i]]
    if len(marked) == 1:
        return marked[0] + 1
    else:
        return 0


def id_boxes_geometry(image, num_cells, lines, dimensions):
    success = False
    # First, select the upper and bottom id lines
    discard = 1 + max([box[1] for box in dimensions])
    lim = 4.5 * lines[-discard][0] - 3.5 * lines[-discard + 1][0]
    hlines = [l for l in lines[:-discard] if l[0] > lim]
    if len(hlines) < 2:
        return None, None
    elif len(hlines) > 2:
        hlines = [hlines[0], hlines[-1]]
    # Check that they are apart enough from the answer tables
    min_height = 0.5 * (lines[-discard + 1][0] - lines[-discard][0])
    if hlines[1][0] - hlines[0][0] < min_height:
        return None, None
    # Now, adjust corners
    pairs_left, pairs_right = line_bounds_adaptive(
        image, hlines[0], hlines[1], images.get_width(image), 5
    )
    all_bounds = [(l[0], r[0], l[1], r[1]) for l in pairs_left for r in pairs_right]
    for bounds in all_bounds[:5]:
        corners = id_boxes_check_points(
            image, bounds, hlines, images.get_width(image), num_cells
        )
        if corners is not None:
            success = True
            break
    if success:
        # Compute the cell of each digit
        id_cells = []
        corners_up, corners_down = corners
        for i in range(0, len(corners_up) - 1):
            cell = capture.CellGeometry(
                corners_up[i],
                corners_up[i + 1],
                corners_down[i],
                corners_down[i + 1],
                None,
                None,
            )
            id_cells.append(cell)
    else:
        id_cells = None
    return hlines, id_cells


def id_boxes_check_points(image, points, hlines, iwidth, num_cells):
    plu, pru, pld, prd = points
    outer_up = [plu, pru]
    outer_down = [pld, prd]
    success = id_boxes_adjust(
        image, outer_up, outer_down, hlines[0], hlines[1], 10, 5, iwidth
    )
    if success:
        corners_up = g.interpolate_line(outer_up[0], outer_up[1], num_cells + 1)
        corners_down = g.interpolate_line(outer_down[0], outer_down[1], num_cells + 1)
        success = id_boxes_adjust(
            image, corners_up, corners_down, hlines[0], hlines[1], 10, 5, iwidth
        )
    if success:
        return (corners_up, corners_down)
    else:
        return None


def id_boxes_adjust(
    image, corners_up, corners_down, line_up, line_down, x_var, rho_var, iwidth
):
    mean_energy = 0.0
    num_corners = len(corners_up)
    for i in range(0, num_corners):
        up = corners_up[i]
        down = corners_down[i]
        selected, energy = id_boxes_adjust_points(
            image, up, down, line_up, line_down, x_var, iwidth
        )
        mean_energy += energy / num_corners
        if energy < param_id_boxes_min_energy_threshold:
            return False
        if selected is not None:
            if rho_var > 0:
                if i == 0:
                    interval = (0, 2 * rho_var)
                elif i == len(corners_up) - 1:
                    interval = (-2 * rho_var, 0)
                else:
                    interval = (-rho_var, rho_var)
                corners_up[i] = id_boxes_adjust_point_vertically(
                    image, selected[0], line_up, interval, iwidth
                )
                corners_down[i] = id_boxes_adjust_point_vertically(
                    image, selected[1], line_down, interval, iwidth
                )
            else:
                corners_up[i] = selected[0]
                corners_down[i] = selected[1]
        else:
            return False
    if mean_energy > param_id_boxes_mean_energy_threshold:
        return True
    else:
        return False


def id_boxes_adjust_points(image, p_up, p_down, line_up, line_down, x_var, iwidth):
    points_up = []
    points_down = []
    for x in range(p_up[0] - x_var, p_up[0] + x_var + 1):
        p = g.line_point(line_up, x=x)
        if p[0] >= 0 and p[0] < iwidth and p[1] >= 0:
            points_up.append(p)
    for x in range(p_down[0] - x_var, p_down[0] + x_var + 1):
        p = g.line_point(line_down, x=x)
        if p[0] >= 0 and p[0] < iwidth and p[1] >= 0:
            points_down.append(p)
    pairs = [(u, v) for u in points_up for v in points_down]
    energies = []
    best = None
    for u, v in pairs:
        energy = id_boxes_match_level(image, u, v)
        if energy > param_id_boxes_energy_break:
            best = ((u, v), energy)
            break
        else:
            energies.append((energy, u, v))
    if best is not None:
        return best
    else:
        energies.sort(reverse=True)
        if len(energies) > 0:
            lim = len(energies)
            for i in range(1, lim):
                if energies[i][0] < energies[0][0]:
                    lim = i
                    break
            best = energies[:lim]
            avgx_up = float(sum([u[0] for (e, u, v) in best])) / len(best)
            avgx_down = float(sum([v[0] for (e, u, v) in best])) / len(best)
            best = [
                (abs(avgx_up - u[0]) + abs(avgx_down - v[0]), u, v)
                for (e, u, v) in best
            ]
            best.sort()
            selected = best[0][1:]
            return selected, energies[0][0]
        else:
            return None, 0.0


def id_boxes_adjust_point_vertically(image, point, line, interval, iwidth):
    rho, theta = line
    lines = [line]
    for i in range(interval[0], interval[1] + 1):
        lines.append((rho + i, theta))
        lines.append((rho - i, theta))
    values = []
    for l in lines:
        match = 0
        for xx in range(point[0] - 2, point[0] + 3):
            x, y = g.line_point(l, x=xx)
            if y >= 0 and x >= 0 and x < iwidth and image[y, x] > 0:
                match += 1
        p = g.line_point(l, x=point[0])
        values.append((match, p[1], p))
    values.sort(reverse=True)
    best = [(m, ppp) for (m, yyy, ppp) in values if m == values[0][0]]
    return best[len(best) // 2][1]


def id_boxes_match_level(image, p0, p1):
    points = [(x, y) for (x, y) in g.walk_line(p0, p1)]
    active = len([(x, y) for (x, y) in points if image[y, x] > 0])
    return float(active) / len(points)


# Utility functions
#
def line_bounds_adaptive(image, line_up, line_down, iwidth, rho_var):
    points_left_up, points_right_up = line_bounds_one_line(
        image, line_up, iwidth, rho_var
    )
    points_left_down, points_right_down = line_bounds_one_line(
        image, line_down, iwidth, rho_var
    )
    pairs_left = [
        line_bounds_rank(p_up, p_down, line_up, line_down)
        for p_up in points_left_up
        for p_down in points_left_down
    ]
    pairs_right = [
        line_bounds_rank(p_up, p_down, line_up, line_down)
        for p_up in points_right_up
        for p_down in points_right_down
    ]
    pairs_left.sort()
    pairs_right.sort()
    return (
        [
            (pair[1], pair[2])
            for pair in pairs_left
            if pair[0] <= param_id_boxes_discard_distance
        ],
        [
            (pair[1], pair[2])
            for pair in pairs_right
            if pair[0] <= param_id_boxes_discard_distance
        ],
    )


def line_bounds_rank(p_up, p_down, line_up, line_down):
    p_down_ideal = g.project_point(p_up, line_up, line_down)
    rank = g.distance(p_down, p_down_ideal)
    return (rank, p_up, p_down)


def line_bounds_one_line(image, line, iwidth, rho_var):
    rho, theta = line
    lines = [line]
    for i in range(1, rho_var + 1):
        lines.append((rho + i, theta))
        lines.append((rho - i, theta))
    points_left = []
    points_right = []
    for l in lines:
        pl, pr = line_bounds(image, l, iwidth)
        if pl is not None:
            points_left.append(pl)
            points_right.append(pr)
    return points_left, points_right


def line_bounds(image, line, iwidth):
    # points of intersection with x = 0 and x = width - 1
    p0 = g.line_point(line, x=0)
    if p0[1] < 0:
        p0 = g.line_point(line, y=0)
    p1 = g.line_point(line, x=iwidth - 1)
    if p1[1] < 0:
        p1 = g.line_point(line, y=0)
    image_dimensions = (images.get_width(image), images.get_height(image))
    if not g.point_is_valid(p0, image_dimensions) or not g.point_is_valid(
        p1, image_dimensions
    ):
        return None, None
    # get bounds
    ini_found = False
    ini = None
    end = None
    last = 0
    count = 0
    for x, y in g.walk_line(p0, p1):
        value = 1 if image[y, x] > 0 else 0
        if value == last:
            count += 1
        else:
            last = value
            count = 1
        if not ini_found:
            if last == 1:
                if count == 1:
                    ini = (x, y)
                elif count == 3:
                    ini_found = True
        else:
            if last == 1 and count > 2:
                end = (x, y)
    if ini is None or end is None:
        ini = None
        end = None
    return ini, end


def process_box_corners(points, dimensions):
    num_boxes = len(dimensions)
    points = sorted(points)
    group1 = [points[0]]
    group2 = []
    # First, look for the other left-most point that is opposite to points[0]
    for i in range(1, len(points)):
        if g.points_closer_to_horizontal(points[0], points[i]):
            group1.append(points[i])
        else:
            group2.append(points[i])
            break
    vertical = g.diff_points(group2[0], group1[0])
    # Now, classify all the other points
    for i in range(len(group1) + len(group2), len(points)):
        cos1 = abs(g.angle_cosine(vertical, g.diff_points(points[i], group1[-1])))
        cos2 = abs(g.angle_cosine(vertical, g.diff_points(points[i], group2[-1])))
        if cos1 < cos2:
            group1.append(points[i])
        else:
            group2.append(points[i])
    if group1[0][1] > group2[0][1]:
        group1, group2 = group2, group1
    if len(group1) != 2 * num_boxes or len(group2) != 2 * num_boxes:
        return []
    boxes = []
    for i in range(0, num_boxes):
        # each box is represented by its corners in this order:
        # (left-up, right-up, left-bottom, right-bottom)
        boxes.append(
            fix_box_if_needed(
                (group1[2 * i], group1[2 * i + 1], group2[2 * i], group2[2 * i + 1])
            )
        )
    corners = []
    for box_dims, box_corners in zip(dimensions, boxes):
        corners.append(construct_box(box_corners, box_dims[0], box_dims[1]))
    return corners


def construct_box(outer_corners, num_columns, num_rows):
    """Returns the corners of all the cells in a box.

    'outer_corners' is a 4-tuple with the coordinates of the outer
    corners of the box: (left-up, right-up, left-bottom,
    right-bottom).

    """
    plu, pru, pld, prd = outer_corners
    line_up_len = g.distance(plu, pru)
    line_down_len = g.distance(pld, prd)
    line_left_len = g.distance(plu, pld)
    line_right_len = g.distance(pru, prd)
    factor_h = line_down_len / line_up_len
    factor_v = line_right_len / line_left_len
    vert_left = g.interpolate_line_progressive(plu, pld, num_rows + 1, factor_h)
    vert_right = g.interpolate_line_progressive(pru, prd, num_rows + 1, factor_h)
    corners = []
    for i in range(0, num_rows + 1):
        pl = vert_left[i]
        pr = vert_right[i]
        corners.append(
            g.interpolate_line_progressive(pl, pr, num_columns + 1, factor_v)
        )
    return corners


def fix_box_if_needed(box_corners):
    """Due to a bug, sometimes corners were not properly detected.
    This code will be kept for a while. It can be removed later the
    error does not happen anymore."""
    plu, pru, pld, prd = box_corners
    if plu[1] > pld[1]:
        plu, pld = pld, plu
        print("Warning: testing box [lu,ru,ld,rd]", box_corners)
        print(" -> points at the left fixed")
    if pru[1] > prd[1]:
        pru, prd = prd, pru
        print("Warning: testing box [lu,ru,ld,rd]", box_corners)
        print(" -> points at the rigth fixed")
    return (plu, pru, pld, prd)


## Debug the upper case with these points: (70, 102), (297, 270),
##                          (62, 276), (260, 101),
##                          (390, 269), (589, 264), (388, 98), (574, 103)

## More for debugging:
##   Testing box [lu,ru,ld,rd] ((97, 92), (288, 90), (95, 258), (287, 258))
##   Testing box [lu,ru,ld,rd] ((409, 90), (597, 257), (408, 258), (586, 89))



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import argparse

from .. import utils
from .. import detection
from .. import capture


def _cmd_options():
    parser = argparse.ArgumentParser(description="Process a single image from file.")
    parser.add_argument(
        "dimensions", help='Answer box dimensions spec. (e.g. "3,5;3,5")'
    )
    parser.add_argument("image", help="Filename of the image")
    parser.add_argument(
        "-t",
        "--hough-threshold",
        dest="hough_threshold",
        type=int,
        default=200,
        help="Hough threshold",
    )
    parser.add_argument(
        "-l",
        "--draw-lines-to",
        dest="draw_lines_to",
        type=str,
        default=None,
        help=("Write the image with the detected lines " "to the given file"),
    )
    parser.add_argument(
        "-p",
        "--save-image-proc-to",
        dest="image_proc_to",
        type=str,
        default=None,
        help="Write the processed image to the given file",
    )
    parser.add_argument(
        "-i",
        "--id-num-digits",
        dest="id_num_digits",
        type=int,
        default=0,
        help=("Detect student id with the given " "number of digits"),
    )
    return parser.parse_args()


def main():
    args = _cmd_options()
    context = detection.ExamDetectorContext(fixed_hough_threshold=args.hough_threshold)
    options = detection.ExamDetector.get_default_options()
    options["capture-from-file"] = True
    options["capture-raw-file"] = args.image
    if args.draw_lines_to is not None:
        options["show-lines"] = True
    if args.id_num_digits:
        options["read-id"] = True
        options["id-num-digits"] = 9
    dimensions, _ = utils.parse_dimensions(args.dimensions)
    detector = detection.ExamDetector(dimensions, context, options)
    success = detector.detect()
    if success:
        print("Detection succeeded :)")
    else:
        print("Detection failed :(")
        print(detector.status)
    if args.draw_lines_to is not None:
        detector.capture.save_image_drawn(args.draw_lines_to)
    if args.image_proc_to:
        capture.save_image(args.image_proc_to, detector.image_proc)


if __name__ == "__main__":
    main()



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import math

import cv2
import numpy as np

from . import geometry as g


# Main image processing functions on numpy images
#
def get_width(image):
    """Returns the width of the numpy image in pixels."""
    return image.shape[1]


def get_height(image):
    """Returns the height of the numpy image in pixels."""
    return image.shape[0]


def new_image(width, height, num_channels):
    if num_channels == 1:
        image = np.zeros((height, width), np.uint8)
    elif num_channels == 3:
        image = np.zeros((height, width, 3), np.uint8)
    else:
        raise ValueError("Wrong number of channels in _new_image()")
    return image


def zero_image(image):
    image[:, :] = 0


def gray_to_rgb(image):
    return cv2.cvtColor(image, cv2.COLOR_GRAY2RGB)


def rgb_to_gray(image):
    return cv2.cvtColor(image, cv2.COLOR_RGB2GRAY)


# Image reading and writing
#
def load_image_grayscale(filename):
    return load_image(filename, flags=cv2.IMREAD_GRAYSCALE)


def load_image(filename, **kwargs):
    return cv2.imread(filename, **kwargs)


# Drawing functions
#
def draw_line(image, line, color=(0, 0, 255, 0)):
    theta = line[1]
    points = set()
    if math.sin(theta) != 0.0:
        points.add(g.line_point(line, x=0))
        points.add(g.line_point(line, x=get_width(image) - 1))
    if math.cos(theta) != 0.0:
        points.add(g.line_point(line, y=0))
        points.add(g.line_point(line, y=get_height(image) - 1))
    p_draw = [
        p
        for p in points
        if p[0] >= 0
        and p[1] >= 0
        and p[0] < get_width(image)
        and p[1] < get_height(image)
    ]
    if len(p_draw) == 2:
        cv2.line(image, p_draw[0], p_draw[1], color, thickness=1)


def draw_point(image, point, color=(255, 0, 0, 0), radius=2):
    x, y = point
    if x >= 0 and x < get_width(image) and y >= 0 and y < get_height(image):
        cv2.circle(image, point, radius, color, thickness=-1)
    else:
        print("draw_point: bad point (%d, %d)" % (x, y))


def draw_text(image, text, color=(255, 0, 0), position=(10, 30)):
    cv2.putText(
        image, text, position, cv2.FONT_HERSHEY_SIMPLEX, 1.0, color, thickness=3
    )



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#
import os
import collections
import random

import cv2
import numpy as np

from .. import geometry as g


class Sample:
    def __init__(self, corners, image=None, image_filename=None, label=None):
        if image is None and not image_filename:
            raise ValueError("Either image of image_filename are needed")
        if corners.shape != (4, 2):
            raise ValueError("Corners must be a 4x2 matrix")
        self.corners = corners
        self.image_filename = image_filename
        self.label = label
        self._image = image
        self._features = None

    @property
    def image(self):
        if self._image is None:
            self._image = cv2.imread(self.image_filename, 0)
            if self._image is None:
                raise ValueError("Cannot load image: {}".format(self.image_filename))
        return self._image

    def check_label(self, label):
        return self.label == label

    def crop(self):
        min_x = min(self.corners[:, 0])
        max_x = max(self.corners[:, 0])
        min_y = min(self.corners[:, 1])
        max_y = max(self.corners[:, 1])
        new_corners = self.corners - np.array([(min_x, min_y)] * 4)
        new_image = self.image[min_y : max_y + 1, min_x : max_x + 1]
        return Sample(new_corners, image=new_image)


class DigitSampleFromCam(Sample):
    def __init__(self, corners, image):
        corners = adjust_cell_corners(image, corners)
        super().__init__(corners, image=image)


class CrossSampleFromCam(Sample):
    def __init__(self, corners, image):
        corners = self._adjust_cell_corners(image, corners)
        super().__init__(corners, image=image)

    @staticmethod
    def _adjust_cell_corners(image, corners):
        plu, prd = g.closer_points_rel(corners[0, :], corners[3, :], 0.8)
        pru, pld = g.closer_points_rel(corners[1, :], corners[2, :], 0.8)
        return np.array((plu, pru, pld, prd))


class SampleSet:
    def __init__(self):
        self.samples_dict = collections.defaultdict(list)

    def __len__(self):
        return sum(len(self.samples_dict[label]) for label in self.samples_dict)

    def __iter__(self):
        return self.iterate_samples()

    @property
    def distribution(self):
        return [(label, len(self.samples_dict[label])) for label in self.samples_dict]

    def load_from_loader(self, loader):
        self.load_from_samples(loader.iterate_samples())

    def load_from_samples(self, samples):
        for sample in samples:
            if sample.label is None:
                raise ValueError("Unlabelled sample in SampleSet")
            self.samples_dict[sample.label].append(sample)

    def load_from_sample_set(self, sample_set):
        for sample in sample_set.iterate_samples():
            if sample.label is None:
                raise ValueError("Unlabelled sample in SampleSet")
            self.samples_dict[sample.label].append(sample)

    def load_from_sample_sets(self, sample_sets):
        for sample_set in sample_sets:
            self.load_from_sample_set(sample_set)

    def samples(self, oversampling=False, downsampling=False):
        return [
            sample
            for sample in self.iterate_samples(
                oversampling=oversampling, downsampling=downsampling
            )
        ]

    def iterate_samples(self, oversampling=False, downsampling=False):
        if oversampling and downsampling:
            raise ValueError("Over and donwsampling are mutually-exclusive")
        if oversampling:
            iterator = self._iterate_samples_oversampling()
        elif downsampling:
            iterator = self._iterate_samples_downsampling()
        else:
            iterator = self._iterate_samples()
        return iterator

    def partition(self, num_groups):
        total_samples = len(self)
        partition_lens = [total_samples // num_groups] * num_groups
        for i in range(total_samples % num_groups):
            partition_lens[i] += 1
        partitions = []
        samples = set(self.samples())
        for partition_len in partition_lens:
            partition = random.sample(samples, partition_len)
            samples -= set(partition)
            sample_set = SampleSet()
            sample_set.load_from_samples(partition)
            partitions.append(sample_set)
        return partitions

    def oversample(self):
        sample_set = SampleSet()
        sample_set.load_from_samples(self.samples(oversampling=True))
        return sample_set

    def downsample(self):
        sample_set = SampleSet()
        sample_set.load_from_samples(self.samples(downsampling=True))
        return sample_set

    def _iterate_samples(self):
        for samples in self.samples_dict.values():
            for sample in samples:
                yield sample

    def _iterate_samples_oversampling(self):
        max_num = self._max_sample_num()
        for samples in self.samples_dict.values():
            rounds = max_num // len(samples)
            remaining = max_num % len(samples)
            for i in range(rounds):
                for sample in samples:
                    yield sample
            for sample in random.sample(samples, remaining):
                yield sample

    def _iterate_samples_downsampling(self):
        min_num = self._min_sample_num()
        for samples in self.samples_dict.values():
            for sample in random.sample(samples, min_num):
                yield sample

    def _max_sample_num(self):
        return max(len(self.samples_dict[label]) for label in self.samples_dict)

    def _min_sample_num(self):
        return min(len(self.samples_dict[label]) for label in self.samples_dict)


class SampleLoader:
    def __init__(self, filename):
        self.filename = filename
        self.dirname = os.path.dirname(filename)

    def samples(self):
        return [sample for sample in self.iterate_samples()]

    def iterate_samples(self):
        with open(self.filename, mode="r") as f:
            for line in f:
                if line.strip():
                    yield self._parse_sample(line)

    def _parse_sample(self, line):
        parts = [p.strip() for p in line.split("\t")]
        if len(parts) != 10:
            raise ValueError("Syntax error in samples file")
        image_path = os.path.join(self.dirname, parts[0])
        label = int(parts[1])
        corners = np.zeros((4, 2), dtype=np.uint16)
        corners[0, 0] = int(parts[2])  # left top
        corners[0, 1] = int(parts[3])
        corners[1, 0] = int(parts[4])  # right top
        corners[1, 1] = int(parts[5])
        corners[2, 0] = int(parts[6])  # left bottom
        corners[2, 1] = int(parts[7])
        corners[3, 0] = int(parts[8])  # right bottom
        corners[3, 1] = int(parts[9])
        return Sample(corners, image_filename=image_path, label=label)


def adjust_cell_corners(image, corners):
    plu = adjust_cell_corner(image, corners[0, :], corners[3, :])
    prd = adjust_cell_corner(image, corners[3, :], corners[0, :])
    pru = adjust_cell_corner(image, corners[1, :], corners[2, :])
    pld = adjust_cell_corner(image, corners[2, :], corners[1, :])
    return np.array([plu, pru, pld, prd])


def adjust_cell_corner(image, corner, towards_corner):
    margin = None
    for x, y in g.walk_line_ordered(corner, towards_corner):
        if margin is None:
            if image[y, x] == 0:
                margin = 2
        else:
            margin -= 1
            if margin == 0:
                return (x, y)
    # In case of failure, return the original point
    return corner
